# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""HTTP utility functions for the reproduce tool."""

import os
import webbrowser

import httplib2

from clusterfuzz._internal.base import json_utils
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.system import shell
from local.butler.reproduce_tool import prompts

GET_METHOD = 'GET'
POST_METHOD = 'POST'

CONFIG_DIRECTORY = os.path.join(
    os.path.expanduser('~'), '.config', 'clusterfuzz')

AUTHORIZATION_CACHE_FILE = os.path.join(CONFIG_DIRECTORY, 'authorization-cache')
AUTHORIZATION_HEADER = 'x-clusterfuzz-authorization'


class SuppressOutput(object):
  """Suppress stdout and stderr.

  We need this to suppress webbrowser's stdout and stderr."""

  def __enter__(self):
    self.stdout = os.dup(1)
    self.stderr = os.dup(2)
    os.close(1)
    os.close(2)
    os.open(os.devnull, os.O_RDWR)

  def __exit__(self, *_):
    os.dup2(self.stdout, 1)
    os.dup2(self.stderr, 2)
    return True


def _get_authorization(force_reauthorization, configuration):
  """Get the value for an oauth authorization header."""
  # Try to read from cache unless we need to reauthorize.
  if not force_reauthorization:
    cached_authorization = utils.read_data_from_file(
        AUTHORIZATION_CACHE_FILE, eval_data=False)
    if cached_authorization:
      return cached_authorization

  # Prompt the user for a code if we don't have one or need a new one.
  oauth_url = configuration.get('oauth_url')
  print('Please login at the following URL to authenticate: {oauth_url}'.format(
      oauth_url=oauth_url))

  with SuppressOutput():
    webbrowser.open(oauth_url, new=1, autoraise=True)

  verification_code = prompts.get_string('Enter verification code')
  return 'VerificationCode {code}'.format(code=verification_code)


def request(url,
            body=None,
            method=POST_METHOD,
            force_reauthorization=False,
            configuration=None):
  """Make an HTTP request to the specified URL."""
  if configuration:
    authorization = _get_authorization(force_reauthorization, configuration)
    headers = {
        'User-Agent': 'clusterfuzz-reproduce',
        'Authorization': authorization
    }
  else:
    headers = {}

  http = httplib2.Http()
  request_body = json_utils.dumps(body) if body is not None else ''
  response, content = http.request(
      url, method=method, headers=headers, body=request_body)

  # If the server returns 401 we may need to reauthenticate. Try the request
  # a second time if this happens.
  if response.status == 401 and not force_reauthorization:
    return request(
        url,
        body,
        method=method,
        force_reauthorization=True,
        configuration=configuration)

  if AUTHORIZATION_HEADER in response:
    shell.create_directory(
        os.path.dirname(AUTHORIZATION_CACHE_FILE), create_intermediates=True)
    utils.write_data_to_file(response[AUTHORIZATION_HEADER],
                             AUTHORIZATION_CACHE_FILE)

  return response, content



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuzzer performance report handler."""

import datetime
import json
import logging
import os
import urllib.parse

from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.google_cloud_utils import big_query
from clusterfuzz._internal.metrics import fuzzer_logs
from clusterfuzz._internal.metrics import fuzzer_stats
from handlers import base_handler
from handlers.performance_report import constants
from handlers.performance_report import performance_analyzer
from libs import access
from libs import handler
from libs import helpers


def _build_rows_and_columns(performance_report):
  """Build table's rows and columns for display."""
  # Build the columns.
  cols = []
  for column in constants.DISPLAY_COLUMNS:
    label = (
        column['title'] +
        '<paper-tooltip>%s</paper-tooltip>' % column['tooltip'])
    cols.append({'label': label, 'type': 'string'})

  # Build the rows.
  rows = []
  for issue in performance_report['issues']:
    rows_data = []
    for column in constants.DISPLAY_COLUMNS:
      rows_data.append({'v': issue[column['name']]})
    rows.append({'c': rows_data})

  table_data = {'cols': cols, 'rows': rows}
  return table_data


def _get_link_html(directory_path, relative_path):
  """Return html link for viewing a particular file on GCS."""
  filename = os.path.basename(relative_path)
  timestamp, extension = os.path.splitext(filename)
  if extension != fuzzer_logs.LOG_EXTENSION:
    return 'Invalid!'

  try:
    # Make sure that timestamp is valid.
    datetime.datetime.strptime(timestamp, fuzzer_logs.TIME_FORMAT)
  except Exception:
    # Invalid timestamp, can't create link. Bail out.
    return 'Invalid!'

  link_path = '%s/%s' % (directory_path, relative_path)
  link_name = filename
  link_url = '/gcs-redirect?%s' % urllib.parse.urlencode({'path': link_path})
  # TODO(mmoroz): build links and other markup things in polymer.
  return '<a href="{link_url}">{link_name}</a>'.format(
      link_url=link_url, link_name=link_name)


def _get_performance_features(fuzzer_name, job_type, datetime_start,
                              datetime_end):
  """Get raw performance features stored in BigQuery."""
  query_fields = [
      fuzzer_stats.QueryField(fuzzer_stats.TestcaseQuery.ALIAS, column, None)
      for column in constants.QUERY_COLUMNS
  ]

  # TODO(mmoroz): the query should be possible for datetime as well object.
  query = fuzzer_stats.TestcaseQuery(
      fuzzer_name=fuzzer_name,
      job_types=[job_type],
      query_fields=query_fields,
      group_by=fuzzer_stats.QueryGroupBy.GROUP_BY_NONE,
      date_start=datetime_start.date(),
      date_end=datetime_end.date())

  client = big_query.Client()

  try:
    result = client.query(query=query.build())
  except Exception as e:
    logging.error('Exception during BigQuery request: %s\n', str(e))
    raise helpers.EarlyExitException('Internal error.', 500)

  if not result.rows:
    raise helpers.EarlyExitException('No stats.', 404)

  return result


def _get_performance_report(fuzzer_name, job_type, performance_report_data):
  """Return performance report."""
  bucket_name = data_handler.get_value_from_job_definition_or_environment(
      job_type, 'FUZZ_LOGS_BUCKET')

  # Load performance data as JSON.
  performance_report = json.loads(performance_report_data)

  # Get logs directory path containing the analyzed logs.
  logs_directory = fuzzer_logs.get_logs_directory(bucket_name, fuzzer_name,
                                                  job_type)

  # Add other display metadata in report.
  for issue in performance_report['issues']:
    # Linkify the examples column.
    # TODO(mmoroz): build this in polymer using dom-repeat.
    issue['examples'] = '<br/>'.join([
        _get_link_html(logs_directory, log_relative_path)
        for log_relative_path in issue['examples']
    ])

    # Add the solutions column explicitly.
    issue['solutions'] = constants.ISSUE_TYPE_SOLUTIONS_MAP[issue['type']]

  return performance_report


def _get_performance_report_data(fuzzer_name, job_type, logs_date):
  """Return performance report data."""
  # Current version works on daily basis the same way as the old version.
  if logs_date == 'latest':
    # Use yesterday's date by UTC to analyze yesterday's fuzzer runs.
    date_start = datetime.datetime.utcnow().date() - datetime.timedelta(days=1)
  else:
    try:
      date_start = datetime.datetime.strptime(logs_date, '%Y-%m-%d').date()
    except ValueError:
      logging.warning('Wrong date format passed to performance report: %s\n',
                      logs_date)
      raise helpers.EarlyExitException('Wrong date format.', 400)

  datetime_start = datetime.datetime.combine(date_start, datetime.time.min)
  datetime_end = datetime_start + datetime.timedelta(days=1)

  features = _get_performance_features(fuzzer_name, job_type, datetime_start,
                                       datetime_end)

  return features, date_start


class Handler(base_handler.Handler):
  """Performance report handler."""

  @handler.get(handler.HTML)
  def get(self, fuzzer_name=None, job_type=None, logs_date=None):
    """Handle a GET request."""
    if not fuzzer_name:
      raise helpers.EarlyExitException('Fuzzer name cannot be empty.', 400)

    if not job_type:
      raise helpers.EarlyExitException('Job type cannot be empty.', 400)

    if not logs_date:
      raise helpers.EarlyExitException('Logs Date cannot be empty.', 400)

    if not access.has_access(fuzzer_name=fuzzer_name, job_type=job_type):
      raise helpers.AccessDeniedException()

    performance_features, date = _get_performance_report_data(
        fuzzer_name, job_type, logs_date)

    performance_data = performance_features.rows

    analyzer = performance_analyzer.LibFuzzerPerformanceAnalyzer()

    # It is possible to break the analysis by requesting outdated stats.
    try:
      total_time = sum(row['actual_duration'] for row in performance_data)
      performance_scores, affected_runs_percents, examples = (
          analyzer.analyze_stats(performance_data))
    except (KeyError, TypeError, ValueError) as e:
      logging.error('Exception during performance analysis: %s\n', str(e))
      raise helpers.EarlyExitException(
          'Cannot analyze performance for the requested time period.', 404)

    # Build performance analysis result.
    performance_issues = analyzer.get_issues(performance_scores,
                                             affected_runs_percents, examples)

    performance_report = performance_analyzer.generate_report(
        performance_issues, fuzzer_name, job_type)

    report = _get_performance_report(fuzzer_name, job_type, performance_report)

    result = {
        'info': {
            'date': str(date),
            'fuzzer_name': report['fuzzer_name'],
            'fuzzer_runs': performance_features.total_count,
            'job_type': report['job_type'],
            'table_data': _build_rows_and_columns(report),
            'total_time': str(datetime.timedelta(seconds=total_time)),
        }
    }
    return self.render('performance-report.html', result)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ML training task."""

import glob
import os
import sys

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.fuzzers.ml.rnn import constants
from clusterfuzz._internal.bot.tasks import ml_train_utils
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.system import shell

# Model script directory.
ML_RNN_SCRIPT_DIR = os.path.join(
    os.path.dirname(__file__), os.pardir, 'fuzzers', 'ml', 'rnn')

# Suffix of temporary directories to place training results.
LOG_DIR_SUFFIX = '_log'
MODEL_DIR_SUFFIX = '_model'
CORPUS_DIR_SUFFIX = '_corpus'

# Training timeout. In theory the longer the model is trained, the more accurate
# it will be.
TRAINING_TIMEOUT = 60 * 60


def get_last_saved_model(model_directory):
  """Get the latest trained model.

  Multiple models may be saved in model_directory. This function will
  find the latest valid model.

  Args:
    model_directory: The directory where models are saved.

  Returns:
    A dictionary with two keys: 'data' and 'index'. Each
    refers to the path of one model file. Empty dictionary will be
    returned if no valid model exists.
  """
  # The dictionary to be returned.
  model_paths = {}

  # Get a list of all index files.
  file_pattern = os.path.join(model_directory,
                              '*' + constants.MODEL_INDEX_SUFFIX)
  index_file_list = list(filter(os.path.isfile, glob.glob(file_pattern)))
  if not index_file_list:
    return model_paths

  # Sort files based on their modification time.
  index_file_list.sort(key=os.path.getmtime, reverse=True)

  # Iterate the list. For each index file, search for corresponding data file.
  for index_file_path in index_file_list:
    index_file_name = os.path.basename(index_file_path)
    file_prefix = os.path.splitext(index_file_name)[0]

    # Find data file with the same prefix (from the same model).
    data_file_name = file_prefix + constants.MODEL_DATA_SUFFIX
    data_file_path = os.path.join(model_directory, data_file_name)

    # Check if they exist.
    if os.path.exists(data_file_path) and os.path.exists(index_file_path):
      model_paths['data'] = data_file_path
      model_paths['index'] = index_file_path
      break

  return model_paths


def get_model_script_path():
  """Get model training script path."""
  return os.path.join(ML_RNN_SCRIPT_DIR, constants.TRAINING_SCRIPT_NAME)


def get_corpus_directory(directory, fuzzer_name):
  """Get corpus directory to place training corpus."""
  return os.path.join(directory, fuzzer_name + CORPUS_DIR_SUFFIX)


def get_model_log_directory(directory, fuzzer_name):
  """Get log directory to keep log files during training."""
  return os.path.join(directory, fuzzer_name + LOG_DIR_SUFFIX)


def get_model_files_directory(directory, fuzzer_name):
  """Get the directory to save model files."""
  return os.path.join(directory, fuzzer_name + MODEL_DIR_SUFFIX)


def upload_model_to_gcs(model_directory, fuzzer_name):
  """Upload the latest model to GCS bucket.

  There might be multiple intermediate models saved during training. This
  function will upload the latest one to GCS bucket.

  Args:
    model_directory: The directory to save intermediate models during training.
    fuzzer_name: The fuzzer the model is trained for.
  """
  # Get latest valid model.
  model_paths = get_last_saved_model(model_directory)
  if not model_paths:
    logs.log_error('No valid RNN model is saved during training.')
    return

  latest_data_file = model_paths['data']
  latest_index_file = model_paths['index']

  # Get GCS model path.
  gcs_model_directory = ml_train_utils.get_gcs_model_directory(
      constants.RNN_MODEL_NAME, fuzzer_name)
  if not gcs_model_directory:
    logs.log_error('Failed to upload model: cannot get GCS model bucket.')
    return

  # Basename of model files.
  data_file_name = constants.RNN_MODEL_NAME + constants.MODEL_DATA_SUFFIX
  index_file_name = constants.RNN_MODEL_NAME + constants.MODEL_INDEX_SUFFIX

  gcs_data_path = '%s/%s' % (gcs_model_directory, data_file_name)
  gcs_index_path = '%s/%s' % (gcs_model_directory, index_file_name)

  logs.log('Uploading the model for %s: %s, %s.' % (fuzzer_name, data_file_name,
                                                    index_file_name))

  # Upload files to GCS.
  result = (
      storage.copy_file_to(latest_data_file, gcs_data_path) and
      storage.copy_file_to(latest_index_file, gcs_index_path))

  if result:
    logs.log('Uploaded ML RNN model for fuzzer %s.' % fuzzer_name)
  else:
    logs.log_error('Failed to upload ML RNN model for fuzzer %s.' % fuzzer_name)


def train_rnn(input_directory,
              model_directory,
              log_directory,
              batch_size=None,
              hidden_state_size=None,
              hidden_layer_size=None):
  """Train ML RNN model.

  Args:
    input_directory: Corpus directory. Required argument for training script.
    model_directory: The directory to save models. Required argument for
        training script.
    log_directory: The directory to keep logs. Required argument for training
        script.
    batch_size: Batch size in each loop.
    hidden_state_size: Hidden state size of LSTM cell.
    hidden_layer_size: Hidden layer size of LSTM model.

  Returns:
    Training result. An object of class `new_process.ProcessResult`.
  """
  # Get the script path to run the model.
  script_path = get_model_script_path()

  # Wrap command and arguments to run training script.
  args_list = [
      script_path,
      constants.INPUT_DIR_ARGUMENT_PREFIX + input_directory,
      constants.MODEL_DIR_ARGUMENT_PREFIX + model_directory,
      constants.LOG_DIR_ARGUMENT_PREFIX + log_directory,
  ]

  # Optional argument.
  if batch_size:
    args_list.append(constants.BATCH_SIZE_ARGUMENT_PREFIX + str(batch_size))
  if hidden_state_size:
    args_list.append(constants.HIDDEN_STATE_ARGUMENT_PREFIX +
                     str(hidden_state_size))
  if hidden_layer_size:
    args_list.append(constants.HIDDEN_LAYER_ARGUMENT_PREFIX +
                     str(hidden_layer_size))

  script_environment = os.environ.copy()

  logs.log('Launching the training with the following arguments: "%s".' %
           str(args_list))

  # Run process in rnn directory.
  rnn_trainer = new_process.ProcessRunner(sys.executable)

  return rnn_trainer.run_and_wait(
      args_list,
      cwd=ML_RNN_SCRIPT_DIR,
      env=script_environment,
      timeout=TRAINING_TIMEOUT)


def execute_task(full_fuzzer_name, job_type):
  """Execute ML RNN training task.

  The task is training RNN model by default. If more models are developed,
  arguments can be modified to specify which model to use.

  Args:
    fuzzer_name: Name of fuzzer, e.g. libpng_read_fuzzer.
    job_type: Job type, e.g. libfuzzer_chrome_asan.
  """
  del job_type

  # Sets up fuzzer binary build.
  fuzz_target = data_handler.get_fuzz_target(full_fuzzer_name)
  if not fuzz_target:
    logs.log_warn(f'Fuzzer not found: {full_fuzzer_name}, skip RNN training.')
    return
  fuzzer_name = fuzz_target.project_qualified_name()

  # Directory to place training files, such as logs, models, corpus.
  # Use |FUZZ_INPUTS_DISK| since it is not size constrained.
  temp_directory = environment.get_value('FUZZ_INPUTS_DISK')

  # Get corpus.
  corpus_directory = get_corpus_directory(temp_directory, fuzzer_name)
  shell.remove_directory(corpus_directory, recreate=True)

  logs.log('Downloading corpus backup for %s.' % fuzzer_name)

  if not ml_train_utils.get_corpus(corpus_directory, fuzzer_name):
    logs.log_error('Failed to download corpus backup for %s.' % fuzzer_name)
    return

  # Get the directory to save models.
  model_directory = get_model_files_directory(temp_directory, fuzzer_name)
  shell.remove_directory(model_directory, recreate=True)

  # Get the directory to save training logs.
  log_directory = get_model_log_directory(temp_directory, fuzzer_name)
  shell.remove_directory(log_directory, recreate=True)

  result = train_rnn(corpus_directory, model_directory, log_directory)

  # Training process exited abnormally but not caused by timeout, meaning
  # error occurred during execution.
  if result.return_code and not result.timed_out:
    if result.return_code == constants.ExitCode.CORPUS_TOO_SMALL:
      logs.log_warn(
          'ML RNN training task for fuzzer %s aborted due to small corpus.' %
          fuzzer_name)
    else:
      logs.log_error(
          'ML RNN training task for fuzzer %s failed with ExitCode = %d.' %
          (fuzzer_name, result.return_code),
          output=utils.decode_to_unicode(result.output))
    return

  # Timing out may be caused by large training corpus, but intermediate models
  # are frequently saved and can be uploaded.
  if result.timed_out:
    logs.log_warn('ML RNN training task for %s timed out.' % fuzzer_name)

  upload_model_to_gcs(model_directory, fuzzer_name)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""NDB utilities. Provides utility functions for NDB."""

from google.cloud import ndb

_GET_BATCH_SIZE = 1000
_MODIFY_BATCH_SIZE = 500


def is_true(boolean_prop):
  """Helper for boolean property filters to avoid lint errors."""
  return boolean_prop == True  # pylint: disable=g-explicit-bool-comparison,singleton-comparison


def is_false(boolean_prop):
  """Helper for boolean property filters to avoid lint errors."""
  return boolean_prop == False  # pylint: disable=g-explicit-bool-comparison,singleton-comparison


def get_all_from_model(model):
  """Get all results from a ndb.Model."""
  return get_all_from_query(model.query())


def get_all_from_query(query, **kwargs):
  """Return all entities based on the query by paging, to avoid query
  expirations on App Engine."""
  # TODO(ochang): Queries no longer expire with new NDB. Remove this and all
  # fix up callers.
  kwargs.pop('batch_size', None)  # No longer supported.
  for entity in query.iter(**kwargs):
    yield entity


def _gen_chunks(values, size):
  """Generate chunks of iterable."""
  values = list(values)
  for i in range(0, len(values), size):
    yield values[i:i + size]


def get_multi(keys):
  """Get multiple entities, working around a limitation in the NDB library with
  the maximum number of keys allowed."""
  result = []
  for chunk in _gen_chunks(keys, _GET_BATCH_SIZE):
    result.extend(ndb.get_multi(chunk))

  return result


def put_multi(entities):
  """Put multiple entities, working around a limitation in the NDB library with
  the maximum number of keys allowed."""
  result = []
  for chunk in _gen_chunks(entities, _MODIFY_BATCH_SIZE):
    result.extend(ndb.put_multi(chunk))

  return result


def delete_multi(keys):
  """Delete multiple entities, working around a limitation in the NDB library
  with the maximum number of keys allowed."""
  for chunk in _gen_chunks(keys, _MODIFY_BATCH_SIZE):
    ndb.delete_multi(chunk)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library to manage bots using GCE instance templates and groups."""

import time

import google_auth_httplib2
import googleapiclient
from googleapiclient.discovery import build
import httplib2

from clusterfuzz._internal.base import retry
from clusterfuzz._internal.google_cloud_utils import credentials

RETRY_COUNT = 8
RETRY_DELAY = 4

REQUEST_TIMEOUT = 180

_SCOPES = [
    'https://www.googleapis.com/auth/cloud-platform',
]


class BotManagerException(Exception):
  """Base exception class."""


class OperationError(BotManagerException):
  """Errors during an operation."""


class RequestError(BotManagerException):
  """Errors during a request."""


class NotFoundError(RequestError):
  """Not found."""


class AlreadyExistsError(RequestError):
  """Already exists."""


class RetryableError(RequestError):
  """Retryable request error."""


class BotManager(object):
  """Manager for bots."""

  def __init__(self, project_id, zone):
    self.project_id = project_id
    self.zone = zone

    creds = credentials.get_default(scopes=_SCOPES)[0]
    http = google_auth_httplib2.AuthorizedHttp(
        creds, http=httplib2.Http(timeout=REQUEST_TIMEOUT))

    self.compute = build('compute', 'v1', http=http, cache_discovery=False)

  def instance_group(self, name):
    """Get an InstanceGroup resource with the given name."""
    return InstanceGroup(name, self)

  def instance_template(self, name):
    """Get an InstanceTemplate resource with the given name."""
    return InstanceTemplate(name, self)


class Resource(object):
  """Represents a resource."""

  _OPERATION_POLL_SECONDS = 5

  def __init__(self, name, manager):
    self.name = name
    self.manager = manager

  @property
  def compute(self):
    return self.manager.compute

  @property
  def project_id(self):
    return self.manager.project_id

  @property
  def zone(self):
    return self.manager.zone

  def get(self):
    raise NotImplementedError

  def exists(self):
    """Return whether or not the resource exists."""
    try:
      self.get()
      return True
    except NotFoundError:
      return False

  def _wait_for_operation(self, operation):
    """Wait for an operation to complete."""
    while True:
      if operation['status'] == 'DONE':
        if 'error' in operation:
          raise OperationError(operation['error'])

        return operation

      time.sleep(self._OPERATION_POLL_SECONDS)

      if 'zone' in operation:
        operation = self.compute.zoneOperations().get(
            project=self.project_id,
            zone=self.zone,
            operation=operation['name']).execute()
      else:
        operation = self.compute.globalOperations().get(
            project=self.project_id, operation=operation['name']).execute()

  def _identity(self, response):
    """Identify function for convenience."""
    return response

  @retry.wrap(
      RETRY_COUNT,
      RETRY_DELAY,
      'handlers.cron.helpers.bot_manager.Resource.execute',
      exception_type=RetryableError)
  def execute(self, request, result_proc=None):
    """Execute a request."""
    if result_proc is None:
      result_proc = self._wait_for_operation

    try:
      response = request.execute()
    except googleapiclient.errors.HttpError as e:
      if e.resp.status in [400, 403, 500, 503]:
        raise RetryableError(str(e))
      if e.resp.status == 404:
        raise NotFoundError(str(e))
      if e.resp.status == 409:
        raise AlreadyExistsError(str(e))

      raise RequestError(str(e))

    return result_proc(response)


class InstanceGroup(Resource):
  """Instance group."""

  # At least 80% of the instances should've been created. Some errors may be
  # expected because of limited resources in the zone.
  MIN_INSTANCES_RATIO = 0.8
  MAX_ERROR_RATIO = 1.0 - MIN_INSTANCES_RATIO

  def _wait_for_instances(self):
    """Wait for instance actions to complete."""
    while True:
      num_instances = 0
      instances_ready = 0
      errors = []

      for instance in self.list_managed_instances():
        num_instances += 1

        if instance['currentAction'] == 'NONE':
          instances_ready += 1
        elif 'lastAttempt' in instance and 'errors' in instance['lastAttempt']:
          errors.append(instance['lastAttempt']['errors'])

      if instances_ready >= max(1, num_instances * self.MIN_INSTANCES_RATIO):
        return

      if len(errors) > num_instances * self.MAX_ERROR_RATIO:
        raise OperationError(errors)

      time.sleep(1)

  def _handle_size_change(self, response):
    """Response handler for operations that change instances."""
    self._wait_for_operation(response)
    self._wait_for_instances()

  def get(self):
    """Get an instance group for a cluster."""
    return self.execute(
        self.compute.instanceGroupManagers().get(
            project=self.project_id,
            zone=self.zone,
            instanceGroupManager=self.name),
        result_proc=self._identity)

  def list_managed_instances(self, instance_filter=None):
    """List managed instances in the group."""
    next_page_token = None

    while True:
      response = self.execute(
          self.compute.instanceGroupManagers().listManagedInstances(
              project=self.project_id,
              zone=self.zone,
              instanceGroupManager=self.name,
              pageToken=next_page_token,
              filter=instance_filter),
          result_proc=self._identity)

      for instance in response['managedInstances']:
        if instance['currentAction'] != 'DELETING':
          # Instances can be stuck in DELETING, don't include them.
          yield instance

      if 'nextPageToken' in response:
        next_page_token = response['nextPageToken']
      else:
        break

  def create(self,
             base_instance_name,
             instance_template,
             size=0,
             wait_for_instances=True):
    """Create this instance group."""
    manager_body = {
        'baseInstanceName': base_instance_name,
        'instanceTemplate': 'global/instanceTemplates/' + instance_template,
        'name': self.name,
        'targetSize': size,
    }

    result_proc = None
    if wait_for_instances:
      result_proc = self._handle_size_change

    self.execute(
        self.compute.instanceGroupManagers().insert(
            project=self.project_id, zone=self.zone, body=manager_body),
        result_proc=result_proc)

  def resize(self, new_size, wait_for_instances=True):
    """Resize this instance group."""
    result_proc = None
    if wait_for_instances:
      result_proc = self._handle_size_change

    self.execute(
        self.compute.instanceGroupManagers().resize(
            project=self.project_id,
            zone=self.zone,
            instanceGroupManager=self.name,
            size=new_size),
        result_proc=result_proc)

  def delete(self):
    """Delete this instance group."""
    self.execute(self.compute.instanceGroupManagers().delete(
        project=self.project_id, zone=self.zone,
        instanceGroupManager=self.name))


class InstanceTemplate(Resource):
  """Instance template."""

  def get(self):
    """Get the instance template."""
    return self.execute(
        self.compute.instanceTemplates().get(
            instanceTemplate=self.name, project=self.project_id),
        result_proc=self._identity)

  def delete(self):
    """Delete the instance template."""
    self.execute(self.compute.instanceTemplates().delete(
        instanceTemplate=self.name, project=self.project_id))

  def create(self, template_body):
    """Create the instance template."""
    template_body['name'] = self.name
    self.execute(self.compute.instanceTemplates().insert(
        project=self.project_id, body=template_body))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gestures related functions."""

import random
import time

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import shell

MAX_CHARS_TO_TYPE = 20
RELOAD_GESTURE = 'key,F5'

COORDINATE_DELTA_MIN = -100
COORDINATE_DELTA_MAX = 200
SCREEN_WIDTH = 1280
SCREEN_HEIGHT = 1024


def _xdotool_path():
  """Return full path to xdotool."""
  return shell.which('xdotool')


def find_windows_for_process(process_id):
  """Return visible windows belonging to a process."""
  pids = utils.get_process_ids(process_id)
  if not pids:
    return []

  xdotool_path = _xdotool_path()
  if not xdotool_path:
    logs.log_error('Xdotool not installed, cannot locate process windows.')
    return []

  visible_windows = []
  for pid in pids:
    windows = (
        shell.execute_command(
            '%s search --all --pid %d --onlyvisible' % (xdotool_path, pid)))

    for line in windows.splitlines():
      if not line.isdigit():
        continue

      visible_windows.append(int(line))

  return visible_windows


def get_random_gestures(gesture_count):
  """Return list of random gesture command strings."""
  gesture_types = [
      'click --repeat TIMES,mbutton',
      'drag',
      'key,ctrl+minus',
      'key,ctrl+plus',
      'key,Function',
      'key,Letter',
      'key,Letters',
      'key,Modifier+Letter',
      'keydown,Letter',
      'keyup,Letter',
      'type,Chars',
      'type,Chars',
      'mousedown,mbutton',
      'mousemove --sync,x y',
      'mousemove_relative --sync,nx ny',
      'mouseup,mbutton',
  ]

  if not random.randint(0, 3):
    gesture_types.append('windowsize,P P')

  gestures = []
  for _ in range(gesture_count):
    random_gesture = utils.random_element_from_list(gesture_types)
    if random_gesture == 'drag':
      gestures.append('mousemove,%d %d' % (random.randint(0, SCREEN_WIDTH),
                                           random.randint(0, SCREEN_HEIGHT)))
      gestures.append('mousedown,1')
      gestures.append('mousemove_relative,0 1')
      gestures.append('mousemove_relative,0 -')
      gestures.append('mousemove,%d %d' % (random.randint(0, SCREEN_WIDTH),
                                           random.randint(0, SCREEN_HEIGHT)))
      gestures.append('mouseup,1')
      continue

    if 'Function' in random_gesture:
      random_gesture = (
          random_gesture.replace('Function', 'F%d' % random.randint(1, 12)))

    elif 'mbutton' in random_gesture:
      random_gesture = random_gesture.replace('TIMES', str(
          random.randint(1, 3)))

      picked_button = 1 if random.randint(0, 4) else random.randint(2, 5)
      random_gesture = random_gesture.replace('mbutton', str(picked_button))

    elif ',x y' in random_gesture:
      random_gesture = random_gesture.replace(
          ',x y', ',%d %d' % (random.randint(0, SCREEN_WIDTH),
                              random.randint(0, SCREEN_HEIGHT)))

    elif ',nx ny' in random_gesture:
      random_gesture = random_gesture.replace(
          ',nx ny', ',%d %d' %
          (random.randint(COORDINATE_DELTA_MIN, COORDINATE_DELTA_MAX),
           random.randint(COORDINATE_DELTA_MIN, COORDINATE_DELTA_MAX)))

    elif ',P P' in random_gesture:
      random_gesture = random_gesture.replace(
          ',P P',
          ',%d%% %d%%' % (random.randint(10, 100), random.randint(10, 100)))

    elif 'Chars' in random_gesture:
      random_gesture = random_gesture.replace('Chars',
                                              "'%s'" % get_text_to_type())

    else:
      if 'Modifier' in random_gesture:
        random_gesture = random_gesture.replace(
            'Modifier',
            utils.random_element_from_list([
                'alt', 'ctrl', 'control', 'meta', 'super', 'shift', 'ctrl+shift'
            ]))

      if 'Letters' in random_gesture:
        num_letters = random.randint(1, 10)
        letters = []
        for _ in range(num_letters):
          letters.append(
              utils.random_element_from_list([
                  'Escape', 'BackSpace', 'Delete', 'Tab', 'space', 'Down',
                  'Return', 'Up', 'Down', 'Left', 'Right',
                  chr(random.randint(48, 57)),
                  chr(random.randint(65, 90)),
                  chr(random.randint(97, 122))
              ]))
        random_gesture = random_gesture.replace('Letters', ' '.join(letters))

      elif 'Letter' in random_gesture:
        random_gesture = random_gesture.replace(
            'Letter',
            utils.random_element_from_list([
                'space',
                chr(random.randint(48, 57)),
                chr(random.randint(65, 90)),
                chr(random.randint(97, 122))
            ]))

        if 'ctrl+c' in random_gesture.lower():
          continue

    gestures.append(random_gesture)

  return gestures


def get_text_to_type():
  """Return text to type."""
  chars = []
  chars_to_type_count = random.randint(1, MAX_CHARS_TO_TYPE)
  meta_chars = [
      '|', '&', ';', '(', ')', '<', '>', ' ', '\t', ',', '\'', '"', '`', '[',
      ']', '{', '}'
  ]

  for _ in range(chars_to_type_count):
    char_code = random.randint(32, 126)
    char = chr(char_code)
    if char in meta_chars:
      continue

    chars.append(char)

  return ''.join(chars)


def run_gestures(gestures, process_id, process_status, start_time, timeout,
                 windows):
  """Run the provided interaction gestures."""
  xdotool_path = _xdotool_path()
  if not xdotool_path:
    logs.log_error('Xdotool not installed, cannot emulate gestures.')
    return

  if not windows:
    windows += find_windows_for_process(process_id)

  for window in windows:
    # Activate the window so that it can receive gestures.
    shell.execute_command(
        '%s windowactivate --sync %d' % (xdotool_path, window))

    for gesture in gestures:
      # If process had exited or our timeout interval has exceeded,
      # just bail out.
      if process_status.finished or time.time() - start_time >= timeout:
        return

      gesture_type, gesture_cmd = gesture.split(',')
      if gesture_type == 'windowsize':
        shell.execute_command(
            '%s %s %d %s' % (xdotool_path, gesture_type, window, gesture_cmd))
      else:
        shell.execute_command(
            '%s %s -- %s' % (xdotool_path, gesture_type, gesture_cmd))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compute Engine helpers."""

import datetime
import random
import time

from googleapiclient.discovery import build
import six

from clusterfuzz._internal.base import dates
from clusterfuzz._internal.base import persistent_cache
from clusterfuzz._internal.metrics import logs

GCE_INSTANCE_INFO_KEY = 'gce_instance_info'
NUM_RETRIES = 10
OPERATION_TIMEOUT = 15 * 60
POLL_INTERVAL = 30
SLEEP_TIME = 10

# TODO(ochang): Allow batching.


def _add_metadata_key_value(items, key, value):
  """Adds a metadata key/value to the "items" list from an instance's
  metadata."""
  replaced_existing = False
  for item in items:
    if item['key'] == key:
      replaced_existing = True
      item['value'] = value
      break

  if not replaced_existing:
    items.append({'key': key, 'value': value})


def _do_operation_with_retries(operation, project, zone, wait_for_completion):
  """Execute an operation, retrying on any exceptions."""
  response = _execute_api_call_with_retries(operation)
  if response is None:
    return False

  if not wait_for_completion:
    # No need to wait, so we are done.
    return True

  for _ in range(NUM_RETRIES + 1):
    try:
      # This could cause exceptions when the response is not ready.
      _wait_for_operation(response, project, zone)
      return True
    except Exception:
      logs.log_error('Failed to wait for Compute Engine operation. '
                     'Original response is %s.' % str(response))
      time.sleep(SLEEP_TIME)
      continue

  return False


def _do_instance_operation(operation, instance_name, project, zone,
                           wait_for_completion):
  """Do a start/reset/stop on the compute engine instance in the given project
  and zone."""
  api = _get_api()
  operation_func = getattr(api.instances(), operation)
  operation = operation_func(instance=instance_name, project=project, zone=zone)

  return _do_operation_with_retries(
      operation, project, zone, wait_for_completion=wait_for_completion)


def _execute_api_call_with_retries(api_func):
  """Execute the given API call, retrying if necessary. Returns the response if
  successful, or None."""
  last_exception = None
  for i in range(NUM_RETRIES + 1):
    try:
      # Try to execute the operation.
      response = api_func.execute()
      last_exception = None
      break
    except Exception as e:
      # Exponential backoff.
      last_exception = str(e)
      sleep_time = random.uniform(1, SLEEP_TIME * (1 << i))
      time.sleep(sleep_time)
      continue

  if last_exception is not None:
    # Failed, log exception with as much information as we can.
    if hasattr(api_func, 'uri'):
      uri = api_func.uri
    else:
      uri = 'unknown'

    logs.log_error('Compute engine API call "%s" failed with exception:\n%s' %
                   (uri, last_exception))
    return None

  return response


def _get_api():
  """Return the compute engine api object."""
  return build('compute', 'v1', cache_discovery=False)


def _get_instance_info(instance_name, project, zone):
  """Return the instance information for a given instance."""
  api = _get_api()
  instance_info_func = api.instances().get(
      instance=instance_name, project=project, zone=zone)
  return _execute_api_call_with_retries(instance_info_func)


def _get_metadata_and_fingerprint(instance_name, project, zone):
  """Return the metadata values and fingerprint for the given instance."""
  instance_info = _get_instance_info(instance_name, project, zone)
  if not instance_info:
    logs.log_error('Failed to fetch instance metadata')
    return None, None

  fingerprint = instance_info['metadata']['fingerprint']
  metadata_items = instance_info['metadata']['items']
  return metadata_items, fingerprint


def _wait_for_operation(response, project, zone):
  """Wait for the given operation to complete."""
  if 'status' in response and response['status'] == 'DONE':
    return

  if 'kind' not in response or response['kind'] != 'compute#operation':
    logs.log_error('Compute api response not an operation.')
    return

  api = _get_api()
  operation = response['name']
  start_time = datetime.datetime.utcnow()

  while not dates.time_has_expired(start_time, seconds=OPERATION_TIMEOUT):
    operation_func = api.zoneOperations().get(
        operation=operation, project=project, zone=zone)
    response = _execute_api_call_with_retries(operation_func)

    if 'status' not in response:
      logs.log_error('Invalid compute engine operation %s.' % str(operation))
      return

    if response['status'] == 'DONE':
      return

    time.sleep(POLL_INTERVAL)

  logs.log_error('Compute engine operation %s timed out.' % str(operation))


def add_metadata(instance_name, project, zone, key, value, wait_for_completion):
  """Add metadata to an existing instance. Replaces existing metadata values
  with the same key."""
  existing_metadata, fingerprint = _get_metadata_and_fingerprint(
      instance_name, project, zone)
  if not existing_metadata:
    return False

  _add_metadata_key_value(existing_metadata, key, value)

  api = _get_api()
  operation = api.instances().setMetadata(
      body={
          'fingerprint': fingerprint,
          'items': existing_metadata,
      },
      instance=instance_name,
      project=project,
      zone=zone)

  return _do_operation_with_retries(operation, project, zone,
                                    wait_for_completion)


def create_disk(disk_name,
                source_image,
                size_gb,
                project,
                zone,
                wait_for_completion=False):
  """Create a disk."""
  api = _get_api()
  operation = api.disks().insert(
      body={
          'name': disk_name,
          'sizeGb': size_gb,
      },
      sourceImage=source_image,
      project=project,
      zone=zone)

  return _do_operation_with_retries(
      operation, project, zone, wait_for_completion=wait_for_completion)


def delete_disk(disk_name, project, zone, wait_for_completion=False):
  """Delete a disk."""
  api = _get_api()
  operation = api.disks().delete(disk=disk_name, project=project, zone=zone)

  return _do_operation_with_retries(
      operation, project, zone, wait_for_completion=wait_for_completion)


def recreate_instance_with_disks(instance_name,
                                 project,
                                 zone,
                                 additional_metadata=None,
                                 wait_for_completion=False):
  """Recreate an instance and its disk."""
  # Get existing instance information.
  # First, try to get instance info from cache.
  # TODO(ochang): Make this more general in case anything else needs to use
  # this method (e.g. appengine).
  instance_info = persistent_cache.get_value(GCE_INSTANCE_INFO_KEY)
  if instance_info is None:
    instance_info = _get_instance_info(instance_name, project, zone)

  # Bail out if we don't have a valid instance information.
  if (not instance_info or 'disks' not in instance_info or
      not instance_info['disks']):
    logs.log_error(
        'Failed to get disk info from existing instance, bailing on instance '
        'recreation.')
    return False

  # Add any additional metadata required for instance booting.
  if additional_metadata:
    for key, value in six.iteritems(additional_metadata):
      items = instance_info.setdefault('metadata', {}).setdefault('items', [])
      _add_metadata_key_value(items, key, value)

  # Cache the latest instance information.
  persistent_cache.set_value(
      GCE_INSTANCE_INFO_KEY, instance_info, persist_across_reboots=True)

  # Delete the instance.
  if not _do_instance_operation(
      'delete', instance_name, project, zone, wait_for_completion=True):
    logs.log_error('Failed to delete instance.')
    return False

  # Get existing disks information, and recreate.
  api = _get_api()
  disks = instance_info['disks']
  for disk in disks:
    disk_source = disk['source']
    disk_name = disk_source.split('/')[-1]

    disk_info_func = api.disks().get(disk=disk_name, project=project, zone=zone)
    disk_info = _execute_api_call_with_retries(disk_info_func)
    if 'sourceImage' not in disk_info or 'sizeGb' not in disk_info:
      logs.log_error(
          'Failed to get source image and size from existing disk, bailing on '
          'instance recreation.')
      return False

    size_gb = disk_info['sizeGb']
    source_image = disk_info['sourceImage']

    # Recreate the disk.
    if not delete_disk(disk_name, project, zone, wait_for_completion=True):
      logs.log_error('Failed to delete disk.')
      return False

    if not create_disk(
        disk_name,
        source_image,
        size_gb,
        project,
        zone,
        wait_for_completion=True):
      logs.log_error('Failed to recreate disk.')
      return False

  # Recreate the instance with the exact same configurations, but not
  # necessarily the same IPs.
  try:
    del instance_info['networkInterfaces'][0]['accessConfigs'][0]['natIP']
  except:
    # This is not a failure. When a bot is stopped, it has no ip/interface.
    pass
  try:
    del instance_info['networkInterfaces'][0]['networkIP']
  except:
    # This is not a failure. When a bot is stopped, it has no ip/interface.
    pass

  operation = api.instances().insert(
      body=instance_info, project=project, zone=zone)

  return _do_operation_with_retries(
      operation, project, zone, wait_for_completion=wait_for_completion)


def remove_metadata(instance_name, project, zone, key, wait_for_completion):
  """Remove a metadata key/value from an existing instance."""
  existing_metadata, fingerprint = _get_metadata_and_fingerprint(
      instance_name, project, zone)
  if not existing_metadata:
    return False

  filtered_metadata = []
  for item in existing_metadata:
    if item['key'] != key:
      filtered_metadata.append(item)

  if len(filtered_metadata) == len(existing_metadata):
    # Nothing to do.
    return True

  api = _get_api()
  operation = api.instances().setMetadata(
      body={
          'fingerprint': fingerprint,
          'items': filtered_metadata,
      },
      instance=instance_name,
      project=project,
      zone=zone)

  return _do_operation_with_retries(operation, project, zone,
                                    wait_for_completion)


def reset_instance(instance_name, project, zone, wait_for_completion=False):
  """Reset an instance."""
  return _do_instance_operation('reset', instance_name, project, zone,
                                wait_for_completion)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script for creating a new deployment config."""

import json
import os
import shutil
import subprocess
import sys

from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient import discovery
import google_auth_httplib2
import httplib2

from local.butler import appengine
from local.butler import common

_REQUIRED_SERVICES = (
    'appengineflex.googleapis.com',
    'bigquery-json.googleapis.com',
    'cloudapis.googleapis.com',
    'cloudbuild.googleapis.com',
    'clouddebugger.googleapis.com',
    'clouderrorreporting.googleapis.com',
    'cloudprofiler.googleapis.com',
    'cloudresourcemanager.googleapis.com',
    'compute.googleapis.com',
    'containerregistry.googleapis.com',
    'datastore.googleapis.com',
    'deploymentmanager.googleapis.com',
    'file.googleapis.com',
    'iam.googleapis.com',
    'iamcredentials.googleapis.com',
    'logging.googleapis.com',
    'monitoring.googleapis.com',
    'oslogin.googleapis.com',
    'pubsub.googleapis.com',
    'redis.googleapis.com',
    'replicapool.googleapis.com',
    'replicapoolupdater.googleapis.com',
    'resourceviews.googleapis.com',
    'siteverification.googleapis.com',
    'sourcerepo.googleapis.com',
    'stackdriver.googleapis.com',
    'storage-api.googleapis.com',
    'storage-component.googleapis.com',
    'vpcaccess.googleapis.com',
)

_NUM_RETRIES = 2
_ENABLE_SERVICE_BATCH_SIZE = 19


class DomainVerifier(object):
  """Domain verifier."""

  def __init__(self, oauth_client_secrets_path):
    flow = InstalledAppFlow.from_client_secrets_file(
        oauth_client_secrets_path,
        scopes=['https://www.googleapis.com/auth/siteverification'])
    credentials = flow.run_console()

    http = google_auth_httplib2.AuthorizedHttp(
        credentials, http=httplib2.Http())

    self.api = discovery.build('siteVerification', 'v1', http=http)

  def get_domain_verification_tag(self, domain):
    """Get the domain verification meta tag."""
    response = self.api.webResource().getToken(
        body={
            'verificationMethod': 'FILE',
            'site': {
                'identifier': domain,
                'type': 'SITE',
            }
        }).execute(num_retries=_NUM_RETRIES)

    return response['token']

  def verify(self, domain):
    """Verify the domain verification meta tag."""
    self.api.webResource().insert(
        body={
            'site': {
                'identifier': domain,
                'type': 'SITE',
            }
        },
        verificationMethod='FILE').execute(num_retries=_NUM_RETRIES)

  def add_owner(self, domain, email):
    """Add a new domain owner."""
    response = self.api.webResource().get(id=domain).execute(
        num_retries=_NUM_RETRIES)

    if email not in response['owners']:
      response['owners'].append(email)

    self.api.webResource().update(
        id=domain, body=response).execute(num_retries=_NUM_RETRIES)


def get_numeric_project_id(gcloud, project_id):
  """Get the numeric project ID."""
  project_info = json.loads(
      gcloud.run('projects', 'describe', project_id, '--format=json'))
  return project_info['projectNumber']


def app_engine_service_account(project_id):
  """Get the default App Engine service account."""
  return project_id + '@appspot.gserviceaccount.com'


def compute_engine_service_account(gcloud, project_id):
  """Get the default compute engine service account."""
  return (get_numeric_project_id(gcloud, project_id) +
          '-compute@developer.gserviceaccount.com')


def enable_services(gcloud):
  """Enable required services."""
  for i in range(0, len(_REQUIRED_SERVICES), _ENABLE_SERVICE_BATCH_SIZE):
    end = i + _ENABLE_SERVICE_BATCH_SIZE
    gcloud.run('services', 'enable', *_REQUIRED_SERVICES[i:i + end])


def replace_file_contents(file_path, replacements):
  """Replace contents of a file."""
  with open(file_path) as f:
    old_contents = f.read()
    contents = old_contents
    for find, replace in replacements:
      contents = contents.replace(find, replace)

  if contents == old_contents:
    return

  with open(file_path, 'w') as f:
    f.write(contents)


def project_bucket(project_id, bucket_name):
  """Return a project-specific bucket name."""
  return '{name}.{project_id}.appspot.com'.format(
      name=bucket_name, project_id=project_id)


def create_new_config(gcloud, project_id, new_config_dir,
                      domain_verification_tag, bucket_replacements,
                      gae_location, gce_zone, firebase_api_key):
  """Create a new config directory."""
  if os.path.exists(new_config_dir):
    print('Overwriting existing directory.')
    shutil.rmtree(new_config_dir)

  gae_region = appengine.region_from_location(gae_location)
  replacements = [
      ('test-clusterfuzz-service-account-email',
       compute_engine_service_account(gcloud, project_id)),
      ('test-clusterfuzz', project_id),
      ('test-project', project_id),
      ('domain-verification-tag', domain_verification_tag),
      ('gae-region', gae_region),
      ('gce-zone', gce_zone),
      ('firebase-api-key', firebase_api_key),
  ]
  replacements.extend(bucket_replacements)

  shutil.copytree(os.path.join('configs', 'test'), new_config_dir)
  for root_dir, _, filenames in os.walk(new_config_dir):
    for filename in filenames:
      file_path = os.path.join(root_dir, filename)
      replace_file_contents(file_path, replacements)


def deploy_appengine(gcloud, config_dir, appengine_location):
  """Deploy to App Engine."""
  try:
    gcloud.run('app', 'describe')
  except common.GcloudError:
    # Create new App Engine app if it does not exist.
    gcloud.run('app', 'create', '--region=' + appengine_location)

  subprocess.check_call([
      'python', 'butler.py', 'deploy', '--force', '--targets', 'appengine',
      '--prod', '--config-dir', config_dir
  ])


def deploy_zips(config_dir):
  """Deploy source zips."""
  subprocess.check_call([
      'python', 'butler.py', 'deploy', '--force', '--targets', 'zips', '--prod',
      '--config-dir', config_dir
  ])


def create_buckets(project_id, buckets):
  """Create buckets."""
  gsutil = common.Gsutil()
  for bucket in buckets:
    try:
      gsutil.run('defstorageclass', 'get', 'gs://' + bucket)
    except common.GsutilError:
      # Create the bucket if it does not exist.
      gsutil.run('mb', '-p', project_id, 'gs://' + bucket)


def set_cors(config_dir, buckets):
  """Sets cors settings."""
  gsutil = common.Gsutil()
  cors_file_path = os.path.join(config_dir, 'gae', 'cors.json')
  for bucket in buckets:
    gsutil.run('cors', 'set', cors_file_path, 'gs://' + bucket)


def add_service_account_role(gcloud, project_id, service_account, role):
  """Add an IAM role to a service account."""
  gcloud.run('projects', 'add-iam-policy-binding', project_id, '--member',
             'serviceAccount:' + service_account, '--role', role)


def execute(args):
  """Create a new config directory and deployment."""
  # Check this early on, as the deployment at the end would fail otherwise.
  if common.is_git_dirty():
    print('Your checkout contains uncommitted changes. Cannot proceed.')
    sys.exit(1)
  verifier = DomainVerifier(args.oauth_client_secrets_path)

  gcloud = common.Gcloud(args.project_id)
  enable_services(gcloud)

  # Get tag for domain verification.
  appspot_domain = 'https://' + args.project_id + '.appspot.com/'
  domain_verification_tag = verifier.get_domain_verification_tag(appspot_domain)

  blobs_bucket = project_bucket(args.project_id, 'blobs')
  deployment_bucket = project_bucket(args.project_id, 'deployment')

  bucket_replacements = (
      ('test-blobs-bucket', blobs_bucket),
      ('test-deployment-bucket', deployment_bucket),
      ('test-bigquery-bucket', project_bucket(args.project_id, 'bigquery')),
      ('test-backup-bucket', project_bucket(args.project_id, 'backup')),
      ('test-coverage-bucket', project_bucket(args.project_id, 'coverage')),
      ('test-fuzzer-logs-bucket', project_bucket(args.project_id,
                                                 'fuzzer-logs')),
      ('test-corpus-bucket', project_bucket(args.project_id, 'corpus')),
      ('test-quarantine-bucket', project_bucket(args.project_id, 'quarantine')),
      ('test-shared-corpus-bucket',
       project_bucket(args.project_id, 'shared-corpus')),
      ('test-fuzz-logs-bucket', project_bucket(args.project_id, 'fuzz-logs')),
      ('test-mutator-plugins-bucket',
       project_bucket(args.project_id, 'mutator-plugins')),
  )

  # Write new configs.
  create_new_config(gcloud, args.project_id, args.new_config_dir,
                    domain_verification_tag, bucket_replacements,
                    args.appengine_location, args.gce_zone,
                    args.firebase_api_key)
  prev_dir = os.getcwd()
  os.chdir(args.new_config_dir)

  # Deploy App Engine and finish verification of domain.
  os.chdir(prev_dir)
  deploy_appengine(
      gcloud, args.new_config_dir, appengine_location=args.appengine_location)
  verifier.verify(appspot_domain)

  # App Engine service account requires:
  # - Domain ownership to create domain namespaced GCS buckets
  # - Datastore export permission for periodic backups.
  # - Service account signing permission for GCS uploads.
  service_account = app_engine_service_account(args.project_id)
  verifier.add_owner(appspot_domain, service_account)
  add_service_account_role(gcloud, args.project_id, service_account,
                           'roles/datastore.importExportAdmin')
  add_service_account_role(gcloud, args.project_id, service_account,
                           'roles/iam.serviceAccountTokenCreator')

  # Create buckets now that domain is verified.
  create_buckets(args.project_id, [bucket for _, bucket in bucket_replacements])

  # Set CORS settings on the buckets.
  set_cors(args.new_config_dir, [blobs_bucket])

  # Set deployment bucket for the cloud project.
  gcloud.run('compute', 'project-info', 'add-metadata',
             '--metadata=deployment-bucket=' + deployment_bucket)

  # Deploy source zips.
  deploy_zips(args.new_config_dir)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the Antlr Tokenizer."""

import unittest

from clusterfuzz._internal.bot.tokenizer.antlr_tokenizer import AntlrTokenizer
from clusterfuzz._internal.bot.tokenizer.grammars.JavaScriptLexer import \
    JavaScriptLexer


class AntlrTokenizerTest(unittest.TestCase):
  """Tests for AntlrTokenizer"""

  def test_empty_list_on_empty_data(self):
    """Test Tokenizer works on empty list"""
    tokenizer = AntlrTokenizer(JavaScriptLexer)
    data = b''

    tokens = tokenizer.tokenize(data)

    self.assertEqual(tokens, [])

  def test_tokenize_simple_js_file(self):
    """Test tokenizer works with sample JS"""
    tokenizer = AntlrTokenizer(JavaScriptLexer)
    txt = b"""async function process(array) {
          for await (let i of array) {
              doSomething(i);
            }
          }"""

    tokens = tokenizer.tokenize(txt)
    self.assertEqual(tokens, [
        'async', ' ', 'function', ' ', 'process', '(', 'array', ')', ' ', '{',
        '\n', '          ', 'for', ' ', 'await', ' ', '(', 'let', ' ', 'i', ' ',
        'of', ' ', 'array', ')', ' ', '{', '\n', '              ',
        'doSomething', '(', 'i', ')', ';', '\n', '            ', '}', '\n',
        '          ', '}'
    ])

  def test_combine_same_as_orig(self):
    """Tests the token combiner"""
    tokenizer = AntlrTokenizer(JavaScriptLexer)
    txt = b"""async function process(array) {
          for await (let i of array) {
              doSomething(i);
            }
          }"""

    tokens = tokenizer.tokenize(txt)

    self.assertEqual(tokenizer.combine(tokens), txt)

  def test_tokenizes_malformed_without_error(self):
    """Tests tokenizer doesnt error on garbage input"""
    tokenizer = AntlrTokenizer(JavaScriptLexer)
    txt = b'aasdfj1  1jhsdf9 1 3@ 1 + => adj 193'

    tokens = tokenizer.tokenize(txt)
    self.assertEqual(tokens, [
        'aasdfj1', '  ', '1', 'jhsdf9', ' ', '1', ' ', '3', '@', ' ', '1', ' ',
        '+', ' ', '=>', ' ', 'adj', ' ', '193'
    ])



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for bisection."""
import datetime
import unittest

import mock

from clusterfuzz._internal.base import bisection
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import mock_config
from clusterfuzz._internal.tests.test_libs import test_utils


@test_utils.with_cloud_emulators('datastore')
class RequestBisectionTest(unittest.TestCase):
  """Tests request_bisection."""

  def setUp(self):
    helpers.patch_environ(self)

    helpers.patch(self, [
        'clusterfuzz._internal.build_management.build_manager.get_primary_bucket_path',
        'clusterfuzz._internal.build_management.build_manager.get_revisions_list',
        'clusterfuzz._internal.build_management.revisions.get_component_range_list',
        'clusterfuzz._internal.config.local_config.ProjectConfig',
        'clusterfuzz._internal.google_cloud_utils.blobs.read_key',
        'clusterfuzz._internal.google_cloud_utils.pubsub.PubSubClient.publish',
    ])
    self.mock.ProjectConfig.return_value = mock_config.MockConfig({
        'env': {
            'PROJECT_NAME': 'test-project',
        },
        'bisect_service': {
            'pubsub_topic': '/projects/project/topics/topic',
        }
    })

    data_types.FuzzTarget(
        id='libFuzzer_proj_target',
        engine='libFuzzer',
        project='proj',
        binary='target').put()

    self.testcase = data_types.Testcase(
        timestamp=datetime.datetime(2021, 1, 1),
        crash_type='crash-type',
        crash_state='A\nB\nC',
        security_flag=True,
        bug_information='1337',
        job_type='libfuzzer_asan_proj',
        fuzzer_name='libFuzzer',
        overridden_fuzzer_name='libFuzzer_proj_target',
        regression='123:456',
        fixed='123:456',
        crash_revision=3,
        security_severity=data_types.SecuritySeverity.MEDIUM,
        additional_metadata='{"last_tested_crash_revision": 4}')
    self.testcase.put()

    self.mock.read_key.return_value = b'reproducer'
    self.mock.get_component_range_list.return_value = [
        {
            'link_text': 'old:new',
        },
    ]

  def _test(self, sanitizer, old_commit='old', new_commit='new'):
    """Test task publication."""
    bisection.request_bisection(self.testcase)
    publish_calls = self.mock.publish.call_args_list
    bisect_types = ('regressed', 'fixed')

    self.assertEqual(2, len(publish_calls))
    for bisect_type, publish_call in zip(bisect_types, publish_calls):
      topic = publish_call[0][1]
      message = publish_call[0][2][0]
      self.assertEqual('/projects/project/topics/topic', topic)
      self.assertEqual(b'reproducer', message.data)
      self.assertDictEqual({
          'crash_state': 'A\nB\nC',
          'crash_type': 'crash-type',
          'security': 'True',
          'severity': 'Medium',
          'fuzz_target': 'target',
          'new_commit': new_commit,
          'old_commit': old_commit,
          'project_name': 'proj',
          'sanitizer': sanitizer,
          'testcase_id': '1',
          'issue_id': '1337',
          'type': bisect_type,
          'timestamp': '2021-01-01T00:00:00',
      }, message.attributes)

    testcase = self.testcase.key.get()
    self.assertTrue(testcase.get_metadata('requested_regressed_bisect'))
    self.assertTrue(testcase.get_metadata('requested_fixed_bisect'))

  def test_request_bisection_asan(self):
    """Basic regressed test (asan)."""
    self.testcase.job_type = 'libfuzzer_asan_proj'
    self.testcase.put()
    self._test('address')

  def test_request_bisection_msan(self):
    """Basic regressed test (asan)."""
    self.testcase.job_type = 'libfuzzer_msan_proj'
    self.testcase.put()
    self._test('memory')

  def test_request_bisection_ubsan(self):
    """Basic regressed test (ubsan)."""
    self.testcase.job_type = 'libfuzzer_ubsan_proj'
    self.testcase.put()
    self._test('undefined')

  def test_request_bisection_blackbox(self):
    """Test request bisection for blackbox."""
    self.testcase.job_type = 'blackbox'
    self.testcase.overridden_fuzzer_name = None
    self.testcase.put()
    bisection.request_bisection(self.testcase)
    self.assertEqual(0, self.mock.publish.call_count)

  def test_request_bisection_non_security(self):
    """Test request bisection for non-security testcases."""
    self.testcase.job_type = 'libfuzzer_asan_proj'
    self.testcase.security_flag = False
    self.testcase.put()
    bisection.request_bisection(self.testcase)
    self.assertEqual(0, self.mock.publish.call_count)

  def test_request_bisection_flaky(self):
    """Test request bisection for flaky testcases."""
    self.testcase.job_type = 'libfuzzer_asan_proj'
    self.testcase.one_time_crasher_flag = True
    self.testcase.put()
    bisection.request_bisection(self.testcase)
    self.assertEqual(0, self.mock.publish.call_count)

  def test_request_bisection_no_bug(self):
    """Test request bisection for testcases with no bug attached."""
    self.testcase.job_type = 'libfuzzer_asan_proj'
    self.testcase.bug_information = ''
    self.testcase.put()
    bisection.request_bisection(self.testcase)
    self.assertEqual(0, self.mock.publish.call_count)

  def test_request_bisection_invalid_range(self):
    """Test request bisection for testcases with no bug attached."""
    self.testcase.job_type = 'libfuzzer_asan_proj'
    self.testcase.regression = 'NA'
    self.testcase.fixed = 'NA'
    self.testcase.put()
    bisection.request_bisection(self.testcase)

    publish_calls = self.mock.publish.call_args_list
    self.assertEqual(1, len(publish_calls))

    publish_call = publish_calls[0]
    topic = publish_call[0][1]
    message = publish_call[0][2][0]
    self.assertEqual('/projects/project/topics/topic', topic)
    self.assertEqual(b'', message.data)
    self.assertDictEqual({
        'testcase_id': '1',
        'type': 'invalid',
    }, message.attributes)

  def test_request_bisection_once_only(self):
    """Test request bisection for testcases isn't repeated if already
    requested."""
    self.testcase.set_metadata('requested_regressed_bisect', True)
    self.testcase.set_metadata('requested_fixed_bisect', True)
    self.testcase.put()
    bisection.request_bisection(self.testcase)
    self.assertEqual(0, self.mock.publish.call_count)

  def test_request_single_commit_range(self):
    """Request bisection with a single commit (invalid range)."""
    self.mock.get_primary_bucket_path.return_value = 'bucket'
    self.mock.get_revisions_list.return_value = list(range(6))
    self.mock.get_component_range_list.return_value = [
        {
            'link_text': 'one',
        },
    ]
    bisection.request_bisection(self.testcase)
    self._test('address', old_commit='one', new_commit='one')
    self.mock.get_component_range_list.assert_has_calls([
        mock.call(123, 456, 'libfuzzer_asan_proj'),
        mock.call(0, 3, 'libfuzzer_asan_proj'),
        mock.call(123, 456, 'libfuzzer_asan_proj'),
        mock.call(4, 5, 'libfuzzer_asan_proj'),
    ])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Monitoring."""
# pylint: disable=invalid-name
# TODO(ochang): Remove V3 from names once all metrics are migrated to
# stackdriver.

import bisect
import collections
import functools
import itertools
import re
import threading
import time

import six

try:
  from google.cloud import monitoring_v3
except (ImportError, RuntimeError):
  monitoring_v3 = None

from google.api_core import exceptions
from google.api_core import retry

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.google_cloud_utils import compute_metadata
from clusterfuzz._internal.google_cloud_utils import credentials
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

CUSTOM_METRIC_PREFIX = 'custom.googleapis.com/'
FLUSH_INTERVAL_SECONDS = 10 * 60  # 10 minutes.
RETRY_DEADLINE_SECONDS = 5 * 60  # 5 minutes.
INITIAL_DELAY_SECONDS = 16
MAXIMUM_DELAY_SECONDS = 2 * 60  # 2 minutes.
MAX_TIME_SERIES_PER_CALL = 200

_retry_wrap = retry.Retry(
    predicate=retry.if_exception_type((
        exceptions.Aborted,
        exceptions.DeadlineExceeded,
        exceptions.ResourceExhausted,
        exceptions.ServerError,
        exceptions.ServiceUnavailable,
    )),
    initial=INITIAL_DELAY_SECONDS,
    maximum=MAXIMUM_DELAY_SECONDS,
    deadline=RETRY_DEADLINE_SECONDS)


class _MockMetric(object):
  """Mock metric object, used for when monitoring isn't available."""

  def _mock_method(self, *args, **kwargs):  # pylint: disable=unused-argument
    pass

  def __getattr__(self, _):
    return self._mock_method


class _FlusherThread(threading.Thread):
  """Flusher thread."""

  def __init__(self):
    super(_FlusherThread, self).__init__()
    self.daemon = True
    self.stop_event = threading.Event()

  def run(self):
    """Run the flusher thread."""
    create_time_series = _retry_wrap(_monitoring_v3_client.create_time_series)
    project_path = _monitoring_v3_client.project_path(
        utils.get_application_id())

    while True:
      try:
        if self.stop_event.wait(FLUSH_INTERVAL_SECONDS):
          return

        time_series = []
        end_time = time.time()
        for metric, labels, start_time, value in _metrics_store.iter_values():
          if (metric.metric_kind ==
              monitoring_v3.enums.MetricDescriptor.MetricKind.GAUGE):
            start_time = end_time

          series = monitoring_v3.types.TimeSeries()
          metric.monitoring_v3_time_series(series, labels, start_time, end_time,
                                           value)
          time_series.append(series)

          if len(time_series) == MAX_TIME_SERIES_PER_CALL:
            create_time_series(project_path, time_series)
            time_series = []

        if time_series:
          create_time_series(project_path, time_series)
      except Exception:
        logs.log_error('Failed to flush metrics.')

  def stop(self):
    self.stop_event.set()
    self.join()


_StoreValue = collections.namedtuple(
    '_StoreValue', ['metric', 'labels', 'start_time', 'value'])


class _MetricsStore(object):
  """In-process metrics store."""

  def __init__(self):
    self._store = {}
    self._lock = threading.RLock()

  def _get_key(self, metric_name, labels):
    """Get the key used for storing values."""
    if labels:
      normalized_labels = tuple(sorted(six.iteritems(labels)))
    else:
      normalized_labels = None

    return (metric_name, normalized_labels)

  def iter_values(self):
    with self._lock:
      for value in six.itervalues(self._store):
        yield value

  def get(self, metric, labels):
    """Get the stored value for the metric."""
    with self._lock:
      key = self._get_key(metric.name, labels)
      return self._store[key]

  def put(self, metric, labels, value):
    """Store new value for the metric."""
    with self._lock:
      key = self._get_key(metric.name, labels)
      if key in self._store:
        start_time = self._store[key].start_time
      else:
        start_time = time.time()

      self._store[key] = _StoreValue(metric, labels, start_time, value)

  def increment(self, metric, labels, delta):
    """Increment a value by |delta|."""
    with self._lock:
      key = self._get_key(metric.name, labels)

      if key in self._store:
        start_time = self._store[key].start_time
        value = self._store[key].value + delta
      else:
        start_time = time.time()
        value = metric.default_value + delta

      self._store[key] = _StoreValue(metric, labels, start_time, value)

  def reset_for_testing(self):
    """Reset all data. Used for tests."""
    with self._lock:
      self._store.clear()


class _Field(object):
  """_Field is the base class used for field specs."""

  def __init__(self, name):
    self.name = name

  @property
  def value_type(self):
    raise NotImplementedError


class StringField(_Field):
  """StringField spec."""

  @property
  def value_type(self):
    return monitoring_v3.enums.LabelDescriptor.ValueType.STRING


class BooleanField(_Field):
  """BooleanField spec."""

  @property
  def value_type(self):
    return monitoring_v3.enums.LabelDescriptor.ValueType.BOOL


class IntegerField(_Field):
  """IntegerField spec."""

  @property
  def value_type(self):
    return monitoring_v3.enums.LabelDescriptor.ValueType.INT64


class Metric(object):
  """Base metric class."""

  def __init__(self, name, description, field_spec):
    self.name = name
    self.description = description
    self.field_spec = field_spec or []

  @property
  def value_type(self):
    raise NotImplementedError

  @property
  def metric_kind(self):
    raise NotImplementedError

  @property
  def default_value(self):
    raise NotImplementedError

  def _set_value(self, point, value):
    raise NotImplementedError

  def get(self, labels=None):
    """Return the current value for the labels. Used for testing."""
    try:
      return _metrics_store.get(self, labels).value
    except KeyError:
      return self.default_value

  def monitoring_v3_metric(self, metric, labels=None):
    """Get the monitoring_v3 Metric."""
    metric.type = CUSTOM_METRIC_PREFIX + self.name

    if not labels:
      return metric

    for key, value in six.iteritems(labels):
      metric.labels[key] = str(value)

    # Default labels.
    bot_name = environment.get_value('BOT_NAME')
    metric.labels['region'] = _get_region(bot_name)

    return metric

  def monitoring_v3_metric_descriptor(self, descriptor):
    """Get the monitoring_v3 MetricDescriptor."""
    descriptor.name = self.name
    descriptor.type = CUSTOM_METRIC_PREFIX + self.name
    descriptor.metric_kind = self.metric_kind
    descriptor.value_type = self.value_type
    descriptor.description = self.description

    for field in itertools.chain(DEFAULT_FIELDS, self.field_spec):
      label_descriptor = descriptor.labels.add()
      label_descriptor.key = field.name
      label_descriptor.value_type = field.value_type

    return descriptor

  def monitoring_v3_time_series(self, time_series, labels, start_time, end_time,
                                value):
    """Get the TimeSeries corresponding to the metric."""
    self.monitoring_v3_metric(time_series.metric, labels)
    time_series.resource.CopyFrom(_monitored_resource)
    time_series.metric_kind = self.metric_kind
    time_series.value_type = self.value_type

    point = time_series.points.add()
    _time_to_timestamp(point.interval.start_time, start_time)
    _time_to_timestamp(point.interval.end_time, end_time)
    self._set_value(point.value, value)

    return time_series


class _CounterMetric(Metric):
  """Counter metric."""

  @property
  def value_type(self):
    return monitoring_v3.enums.MetricDescriptor.ValueType.INT64

  @property
  def metric_kind(self):
    return monitoring_v3.enums.MetricDescriptor.MetricKind.CUMULATIVE

  @property
  def default_value(self):
    return 0

  def increment(self, labels=None):
    self.increment_by(1, labels=labels)

  def increment_by(self, count, labels=None):
    _metrics_store.increment(self, labels, count)

  def _set_value(self, point, value):
    """Get Point."""
    point.int64_value = value


class _GaugeMetric(Metric):
  """Gauge metric."""

  @property
  def value_type(self):
    return monitoring_v3.enums.MetricDescriptor.ValueType.INT64

  @property
  def metric_kind(self):
    return monitoring_v3.enums.MetricDescriptor.MetricKind.GAUGE

  @property
  def default_value(self):
    return 0

  def set(self, value, labels=None):
    _metrics_store.put(self, labels, value)

  def _set_value(self, point, value):
    """Get Point."""
    point.int64_value = value


class _Bucketer(object):
  """Bucketer."""

  def bucket_for_value(self, value):
    """Get the bucket index for the given value."""
    return bisect.bisect(self._lower_bounds, value) - 1

  @property
  def num_buckets(self):
    return len(self._lower_bounds)


class FixedWidthBucketer(_Bucketer):
  """Fixed width bucketer."""

  def __init__(self, width, num_finite_buckets=100):
    self.width = width
    self.num_finite_buckets = num_finite_buckets

    # [-Inf, 0), [0, width), [width, 2*width], ... , [n*width, Inf)
    self._lower_bounds = [float('-Inf')]
    self._lower_bounds.extend(
        [width * i for i in range(num_finite_buckets + 1)])


class GeometricBucketer(_Bucketer):
  """Geometric bucketer."""

  def __init__(self, growth_factor=10**0.2, num_finite_buckets=100, scale=1.0):
    self.growth_factor = growth_factor
    self.num_finite_buckets = num_finite_buckets
    self.scale = scale

    # [-Inf, scale), [scale, scale*growth),
    # [scale*growth^i, scale*growth^(i+1)), ..., [scale*growth^n, Inf)
    self._lower_bounds = [float('-Inf')]
    self._lower_bounds.extend(
        [scale * growth_factor**i for i in range(num_finite_buckets + 1)])


class _Distribution(object):
  """Holds a distribution."""

  def __init__(self, bucketer):
    self.bucketer = bucketer
    self.buckets = [0 for _ in range(bucketer.num_buckets)]
    self.sum = 0
    self.count = 0

  def add(self, value):
    self.buckets[self.bucketer.bucket_for_value(value)] += 1
    self.count += 1
    self.sum += value
    return self

  __add__ = add

  def monitoring_v3_distribution(self, distribution):
    """Set the monitoring_v3 Distribution value."""
    distribution.count = self.count
    if self.count:
      distribution.mean = float(self.sum) / self.count
    else:
      distribution.mean = 0.0

    if isinstance(self.bucketer, FixedWidthBucketer):
      distribution.bucket_options.linear_buckets.offset = 0
      distribution.bucket_options.linear_buckets.width = self.bucketer.width
      distribution.bucket_options.linear_buckets.num_finite_buckets = (
          self.bucketer.num_finite_buckets)
    else:
      assert isinstance(self.bucketer, GeometricBucketer)

      distribution.bucket_options.exponential_buckets.scale = (
          self.bucketer.scale)
      distribution.bucket_options.exponential_buckets.growth_factor = (
          self.bucketer.growth_factor)
      distribution.bucket_options.exponential_buckets.num_finite_buckets = (
          self.bucketer.num_finite_buckets)

    distribution.bucket_counts.extend(self.buckets)


class _CumulativeDistributionMetric(Metric):
  """Cumulative distribution metric."""

  def __init__(self, name, description, bucketer, field_spec=None):
    super(_CumulativeDistributionMetric, self).__init__(
        name, description=description, field_spec=field_spec)
    self.bucketer = bucketer

  @property
  def value_type(self):
    return monitoring_v3.enums.MetricDescriptor.ValueType.DISTRIBUTION

  @property
  def metric_kind(self):
    return monitoring_v3.enums.MetricDescriptor.MetricKind.CUMULATIVE

  @property
  def default_value(self):
    return _Distribution(self.bucketer)

  def add(self, value, labels=None):
    _metrics_store.increment(self, labels, value)

  def _set_value(self, point, value):
    value.monitoring_v3_distribution(point.distribution_value)


# Global state.
_metrics_store = _MetricsStore()
_monitoring_v3_client = None
_flusher_thread = None
_monitored_resource = None

# Add fields very conservatively here. There is a limit of 10 labels per metric
# descriptor, and metrics should be low in cardinality. That is, only add fields
# which have a small number of possible values.
DEFAULT_FIELDS = [
    StringField('region'),
]


def check_module_loaded(module):
  """Used for mocking."""
  return module is not None


def stub_unavailable(module):
  """Decorator to stub out functions on failed imports."""

  def decorator(func):

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
      if check_module_loaded(module):
        return func(*args, **kwargs)

      return _MockMetric()

    return wrapper

  return decorator


def _initialize_monitored_resource():
  """Monitored resources."""
  global _monitored_resource
  _monitored_resource = monitoring_v3.types.MonitoredResource()

  # TODO(ochang): Use generic_node when that is available.
  _monitored_resource.type = 'gce_instance'

  # The project ID must be the same as the one we write metrics to, not the ID
  # where the instance lives.
  _monitored_resource.labels['project_id'] = utils.get_application_id()

  # Use bot name here instance as that's more useful to us.
  _monitored_resource.labels['instance_id'] = environment.get_value('BOT_NAME')

  if compute_metadata.is_gce():
    # Returned in the form projects/{id}/zones/{zone}
    zone = compute_metadata.get('instance/zone').split('/')[-1]
    _monitored_resource.labels['zone'] = zone
  else:
    # Default zone for instances not on GCE.
    _monitored_resource.labels['zone'] = 'us-central1-f'


def _time_to_timestamp(timestamp, time_seconds):
  """Convert result of time.time() to Timestamp."""
  timestamp.seconds = int(time_seconds)
  timestamp.nanos = int((time_seconds - timestamp.seconds) * 10**9)


def initialize():
  """Initialize if monitoring is enabled for this bot."""
  global _monitoring_v3_client
  global _flusher_thread

  if environment.get_value('LOCAL_DEVELOPMENT'):
    return

  if not local_config.ProjectConfig().get('monitoring.enabled'):
    return

  if check_module_loaded(monitoring_v3):
    _initialize_monitored_resource()
    _monitoring_v3_client = monitoring_v3.MetricServiceClient(
        credentials=credentials.get_default()[0])
    _flusher_thread = _FlusherThread()
    _flusher_thread.start()


def stop():
  """Stops monitoring and cleans up (only if monitoring is enabled)."""
  if _flusher_thread:
    _flusher_thread.stop()


def metrics_store():
  """Get the per-process metrics store."""
  return _metrics_store


def _get_region(bot_name):
  """Get bot region."""
  try:
    regions = local_config.MonitoringRegionsConfig()
  except errors.BadConfigError:
    return 'unknown'

  for pattern in regions.get('patterns'):
    if re.match(pattern['pattern'], bot_name):
      return pattern['name']

  return 'unknown'


@stub_unavailable(monitoring_v3)
def CounterMetric(name, description, field_spec):
  """Build _CounterMetric."""
  return _CounterMetric(name, field_spec=field_spec, description=description)


@stub_unavailable(monitoring_v3)
def GaugeMetric(name, description, field_spec):
  """Build _CounterMetric."""
  return _GaugeMetric(name, field_spec=field_spec, description=description)


@stub_unavailable(monitoring_v3)
def CumulativeDistributionMetric(name, description, bucketer, field_spec):
  """Build _CounterMetric."""
  return _CumulativeDistributionMetric(
      name, description=description, bucketer=bucketer, field_spec=field_spec)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Jira issue tracker."""

from dateutil import parser

from clusterfuzz._internal.config import db_config
from libs.issue_management import issue_tracker
from libs.issue_management.jira.issue_tracker_manager import IssueTrackerManager


class Issue(issue_tracker.Issue):
  """Represents an issue."""

  def __init__(self, itm, jira_issue):
    self.itm = itm
    self.jira_issue = jira_issue

    self._ccs = issue_tracker.LabelStore(self.itm.get_watchers(self.jira_issue))
    self._components = issue_tracker.LabelStore(
        self.jira_issue.fields.components)
    self._labels = issue_tracker.LabelStore(self.jira_issue.fields.labels)

  @property
  def issue_tracker(self):
    """The IssueTracker for this issue."""
    return IssueTracker(self.itm)

  @property
  def id(self):
    """The issue identifier."""
    return self.jira_issue.key

  @property
  def title(self):
    """The issue title."""
    return self.jira_issue.fields.summary

  @title.setter
  def title(self, new_title):
    self.jira_issue.fields.summary = new_title

  @property
  def reporter(self):
    """The issue reporter."""
    return self.jira_issue.fields.reporter

  @reporter.setter
  def reporter(self, new_reporter):
    self.jira_issue.fields.reporter = new_reporter

  @property
  def is_open(self):
    """Whether the issue is open."""
    return self.jira_issue.resolution not in ['Closed', 'Done', 'Resolved']

  @property
  def closed_time(self):
    return parser.parse(self.jira_issue.fields.resolutiondate)

  @property
  def status(self):
    """The issue status."""
    return self.jira_issue.fields.status

  @status.setter
  def status(self, new_status):
    self.jira_issue.fields.status = new_status

  @property
  def body(self):
    """The issue body."""
    return self.jira_issue.fields.description

  @body.setter
  def body(self, new_body):
    self.jira_issue.fields.description = new_body

  @property
  def assignee(self):
    """The issue assignee."""
    return self.jira_issue.fields.assignee

  @assignee.setter
  def assignee(self, new_assignee):
    self.jira_issue.fields.assignee = new_assignee

  @property
  def ccs(self):
    """The issue CC list."""
    return self._ccs

  @property
  def labels(self):
    """The issue labels list."""
    return self._labels

  @property
  def components(self):
    """The issue component list."""
    return self._components

  # FIXME: Add support for new_comment and notify arguments
  def save(self, new_comment=None, notify=True):  # pylint: disable=unused-argument
    """Save the issue."""
    for added in self._components.added:
      self.components.add(added)
    for removed in self._components.removed:
      self.components.remove(removed)
    self._components.reset_tracking()

    for added in self._ccs.added:
      self.ccs.add(added)
    for removed in self._ccs.removed:
      self.ccs.remove(removed)
    self._ccs.reset_tracking()

    for added in self._labels.added:
      self.labels.add(added)
    for removed in self._labels.removed:
      self.labels.remove(removed)
    self._labels.reset_tracking()

    self.itm.save(self)

  def actions(self):
    pass

  def merged_into(self):
    pass


class IssueTracker(issue_tracker.IssueTracker):
  """Issue tracker interface."""

  def __init__(self, itm):
    self._itm = itm

  @property
  def project(self):
    return self._itm.project_name

  def new_issue(self):
    jira_issue = self._itm.create()
    return Issue(self._itm, jira_issue)

  def get_issue(self, issue_id):
    jira_issue = self._itm.get_issue(issue_id)
    if not jira_issue:
      return None

    return Issue(self._itm, jira_issue)

  def find_issues(self, keywords=None, only_open=False):
    """Find issues."""
    search_text = 'project = {project_name}' + _get_search_text(keywords)
    search_text.format(project_name=self._itm.project_name)
    if only_open:
      search_text += ' AND resolution = Unresolved'
    issues = self._itm.get_issues(search_text)
    return [Issue(self._itm, issue) for issue in issues]

  def issue_url(self, issue_id):
    """Return the issue URL with the given ID."""
    config = db_config.get()
    url = config.jira_url + '/browse/' + str(issue_id)
    return url

  # FIXME: Add support for keywords and only_open arguments
  def find_issues_url(self, keywords=None, only_open=None):  # pylint: disable=unused-argument
    pass


def _get_issue_tracker_manager_for_project(project_name):
  """Return jira issue tracker manager for the given project."""
  # If there is no issue tracker set, bail out.
  if not project_name or project_name == 'disabled':
    return None

  return IssueTrackerManager(project_name=project_name)


def get_issue_tracker(project_name, config):  # pylint: disable=unused-argument
  """Get the issue tracker for the project name."""
  itm = _get_issue_tracker_manager_for_project(project_name)
  if itm is None:
    return None

  return IssueTracker(itm)


def _get_search_text(keywords):
  """Get search text."""
  search_text = ''
  for keyword in keywords:
    search_text += ' AND text ~ "%s"' % keyword

  return search_text



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for corpus_manager."""

import datetime
import os
import unittest

import mock
from pyfakefs import fake_filesystem_unittest

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.fuzzing import corpus_manager
from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class GcsCorpusTest(unittest.TestCase):
  """GcsCorpus tests."""

  def setUp(self):
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.fuzzing.corpus_manager._count_corpus_files',
        'multiprocessing.cpu_count',
        'subprocess.Popen',
    ])

    self.mock.Popen.return_value.poll.return_value = 0
    self.mock.Popen.return_value.communicate.return_value = (None, None)
    self.mock._count_corpus_files.return_value = 1  # pylint: disable=protected-access

    os.environ['GSUTIL_PATH'] = '/gsutil_path'

  def test_rsync_to_disk(self):
    """Test rsync_to_disk."""
    self.mock.cpu_count.return_value = 1
    corpus = corpus_manager.GcsCorpus('bucket')
    self.assertTrue(corpus.rsync_to_disk('/dir', timeout=60))

    self.assertEqual(self.mock.Popen.call_args[0][0], [
        '/gsutil_path/gsutil', '-m', '-o', 'GSUtil:parallel_thread_count=16',
        '-q', 'rsync', '-r', '-d', 'gs://bucket/', '/dir'
    ])

    self.mock.cpu_count.return_value = 2
    corpus = corpus_manager.GcsCorpus('bucket')
    self.assertTrue(corpus.rsync_to_disk('/dir', timeout=60))
    self.assertEqual(self.mock.Popen.call_args[0][0], [
        '/gsutil_path/gsutil', '-m', '-q', 'rsync', '-r', '-d', 'gs://bucket/',
        '/dir'
    ])

  def test_rsync_from_disk(self):
    """Test rsync_from_disk."""
    self.mock.cpu_count.return_value = 1
    corpus = corpus_manager.GcsCorpus('bucket')
    self.assertTrue(corpus.rsync_from_disk('/dir'))

    self.assertEqual(self.mock.Popen.call_args[0][0], [
        '/gsutil_path/gsutil', '-m', '-o', 'GSUtil:parallel_thread_count=16',
        '-q', 'rsync', '-r', '-d', '/dir', 'gs://bucket/'
    ])

    self.mock.cpu_count.return_value = 2
    corpus = corpus_manager.GcsCorpus('bucket')
    self.assertTrue(corpus.rsync_from_disk('/dir'))
    self.assertEqual(self.mock.Popen.call_args[0][0], [
        '/gsutil_path/gsutil', '-m', '-q', 'rsync', '-r', '-d', '/dir',
        'gs://bucket/'
    ])

  def test_upload_files(self):
    """Test upload_files."""
    mock_popen = self.mock.Popen.return_value

    self.mock.cpu_count.return_value = 1
    corpus = corpus_manager.GcsCorpus('bucket')
    self.assertTrue(corpus.upload_files(['/dir/a', '/dir/b']))

    self.assertEqual(self.mock.Popen.call_args[0][0], [
        '/gsutil_path/gsutil', '-m', '-o', 'GSUtil:parallel_thread_count=16',
        'cp', '-I', 'gs://bucket/'
    ])

    mock_popen.communicate.assert_called_with(b'/dir/a\n/dir/b')

    self.mock.cpu_count.return_value = 2
    corpus = corpus_manager.GcsCorpus('bucket')
    self.assertTrue(corpus.upload_files(['/dir/a', '/dir/b']))
    self.assertEqual(self.mock.Popen.call_args[0][0],
                     ['/gsutil_path/gsutil', '-m', 'cp', '-I', 'gs://bucket/'])


class RsyncErrorHandlingTest(unittest.TestCase):
  """Rsync error handling tests."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.fuzzing.corpus_manager._count_corpus_files',
        'clusterfuzz._internal.google_cloud_utils.gsutil.GSUtilRunner.run_gsutil',
    ])

  def test_rsync_error_below_threshold(self):
    """Test rsync returning errors (but they're below threshold)."""
    output = (
        b'blah\n'
        b'blah\n'
        b'CommandException: 10 files/objects could not be copied/removed.\n')

    self.mock._count_corpus_files.return_value = 10  # pylint: disable=protected-access
    self.mock.run_gsutil.return_value = new_process.ProcessResult(
        command=['/fake'],
        return_code=1,
        output=output,
        time_executed=10.0,
        timed_out=False,
    )

    corpus = corpus_manager.GcsCorpus('bucket')
    self.assertTrue(corpus.rsync_to_disk('/dir', timeout=60))

    self.mock.run_gsutil.return_value = new_process.ProcessResult(
        command=['/fake'],
        return_code=1,
        output=output,
        time_executed=30.0,
        timed_out=True,
    )
    self.assertFalse(corpus.rsync_to_disk('/dir', timeout=60))

  def test_rsync_error_below_threshold_with_not_found_errors(self):
    """Test rsync returning errors (below threshold, but with not found errors
    and overall error count more than threshold)."""
    output = (
        b'blah\n' + b'[Errno 2] No such file or directory\n' * 10 +
        b'NotFoundException: 404 gs://bucket/file001 does not exist.\n' * 180 +
        b'CommandException: 200 files/objects could not be copied/removed.\n')

    self.mock._count_corpus_files.return_value = 10  # pylint: disable=protected-access
    self.mock.run_gsutil.return_value = new_process.ProcessResult(
        command=['/fake'],
        return_code=1,
        output=output,
        time_executed=10.0,
        timed_out=False,
    )

    corpus = corpus_manager.GcsCorpus('bucket')
    self.assertTrue(corpus.rsync_to_disk('/dir', timeout=60))

    self.mock.run_gsutil.return_value = new_process.ProcessResult(
        command=['/fake'],
        return_code=1,
        output=output,
        time_executed=30.0,
        timed_out=True,
    )
    self.assertFalse(corpus.rsync_to_disk('/dir', timeout=60))

  def test_rsync_error_above_threshold(self):
    """Test rsync returning errors (above threshold)."""
    output = (
        b'blah\n'
        b'blah\n'
        b'CommandException: 11 files/objects could not be copied/removed.\n')

    self.mock.run_gsutil.return_value = new_process.ProcessResult(
        command=['/fake'],
        return_code=1,
        output=output,
        time_executed=10.0,
        timed_out=False,
    )

    corpus = corpus_manager.GcsCorpus('bucket')
    self.assertFalse(corpus.rsync_to_disk('/dir', timeout=60))


class FuzzTargetCorpusTest(fake_filesystem_unittest.TestCase):
  """FuzzTargetCorpus tests."""

  def setUp(self):
    """Setup for fuzz target corpus test."""
    test_helpers.patch_environ(self)

    os.environ['GSUTIL_PATH'] = '/gsutil_path'
    os.environ['CORPUS_BUCKET'] = 'bucket'

    test_helpers.patch(self, [
        'clusterfuzz._internal.fuzzing.corpus_manager._count_corpus_files',
        'multiprocessing.cpu_count',
        'subprocess.Popen',
    ])

    self.mock.Popen.return_value.poll.return_value = 0
    self.mock.Popen.return_value.communicate.return_value = (None, None)
    self.mock.cpu_count.return_value = 2
    self.mock._count_corpus_files.return_value = 1  # pylint: disable=protected-access
    test_utils.set_up_pyfakefs(self)
    self.fs.create_dir('/dir')

  def test_rsync_to_disk(self):
    """Test rsync_to_disk."""
    corpus = corpus_manager.FuzzTargetCorpus('libFuzzer', 'fuzzer')
    self.assertTrue(corpus.rsync_to_disk('/dir', timeout=60))
    self.assertEqual(self.mock.Popen.call_args[0][0], [
        '/gsutil_path/gsutil',
        '-m',
        '-q',
        'rsync',
        '-r',
        '-d',
        'gs://bucket/libFuzzer/fuzzer/',
        '/dir',
    ])

  def test_rsync_to_disk_with_regressions(self):
    """Test rsync_to_disk, with regressions set."""
    corpus = corpus_manager.FuzzTargetCorpus(
        'libFuzzer', 'fuzzer', include_regressions=True)
    self.assertTrue(corpus.rsync_to_disk('/dir', timeout=60))

    commands = [call_arg[0][0] for call_arg in self.mock.Popen.call_args_list]

    self.assertEqual(commands, [
        [
            '/gsutil_path/gsutil',
            '-m',
            '-q',
            'rsync',
            '-r',
            '-d',
            'gs://bucket/libFuzzer/fuzzer/',
            '/dir',
        ],
        [
            '/gsutil_path/gsutil',
            '-m',
            '-q',
            'rsync',
            '-r',
            'gs://bucket/libFuzzer/fuzzer_regressions/',
            '/dir/regressions',
        ],
    ])

  def test_rsync_from_disk(self):
    """Test rsync_from_disk."""
    corpus = corpus_manager.FuzzTargetCorpus('libFuzzer', 'fuzzer')
    self.assertTrue(corpus.rsync_from_disk('/dir'))

    self.assertEqual(self.mock.Popen.call_args[0][0], [
        '/gsutil_path/gsutil', '-m', '-q', 'rsync', '-r', '-d', '/dir',
        'gs://bucket/libFuzzer/fuzzer/'
    ])

  def test_upload_files(self):
    """Test upload_files."""
    mock_popen = self.mock.Popen.return_value

    corpus = corpus_manager.FuzzTargetCorpus('libFuzzer', 'fuzzer')
    self.assertTrue(corpus.upload_files(['/dir/a', '/dir/b']))
    mock_popen.communicate.assert_called_with(b'/dir/a\n/dir/b')

    self.assertEqual(self.mock.Popen.call_args[0][0], [
        '/gsutil_path/gsutil', '-m', 'cp', '-I', 'gs://bucket/libFuzzer/fuzzer/'
    ])


class CorpusBackupTest(fake_filesystem_unittest.TestCase):
  """Corpus backup tests."""

  def _mock_make_archive(self, archive_path, backup_format, _):
    path = archive_path + '.' + backup_format
    self.fs.create_file(path)

    return path

  def setUp(self):
    """Setup for corpus backup test."""
    test_helpers.patch_environ(self)

    test_utils.set_up_pyfakefs(self)
    self.fs.create_dir('/dir')

    os.environ['GSUTIL_PATH'] = '/gsutil_path'
    os.environ['CORPUS_BUCKET'] = 'bucket'

    test_helpers.patch(self, [
        'clusterfuzz._internal.base.utils.utcnow',
        'clusterfuzz._internal.google_cloud_utils.storage.copy_blob',
        'clusterfuzz._internal.google_cloud_utils.storage.copy_file_to',
        'multiprocessing.cpu_count',
        'shutil.make_archive',
    ])

    self.mock.copy_blob.return_value = True
    self.mock.copy_file_to.return_value = True
    self.mock.cpu_count.return_value = 2
    self.mock.make_archive.side_effect = self._mock_make_archive
    self.mock.utcnow.return_value = datetime.datetime(2017, 1, 1)

  def test_backup_corpus(self):
    """Test backup_corpus."""
    libfuzzer_corpus = corpus_manager.FuzzTargetCorpus('libFuzzer', 'fuzzer')

    corpus_manager.backup_corpus('backup_bucket', libfuzzer_corpus, '/dir')

    self.mock.copy_file_to.assert_has_calls([
        mock.call('/2017-01-01.zip',
                  'gs://backup_bucket/corpus/libFuzzer/fuzzer/2017-01-01.zip')
    ])

    self.mock.copy_blob.assert_has_calls([
        mock.call('gs://backup_bucket/corpus/libFuzzer/fuzzer/2017-01-01.zip',
                  'gs://backup_bucket/corpus/libFuzzer/fuzzer/latest.zip'),
    ])


class FileMixin(object):
  """Mixin with a setUp implementation and attributes that are useful for test
  classes dealing with cleaning filenames for Windows."""
  # Make the content greater than chunk size.
  FILE_CONTENTS = 'hello' * 30000
  FILE_SHA1SUM = 'bf720c91eb988c589a486407d7912da17bdb201f'
  DIRECTORY = '/corpus'
  FILE_PATH = os.path.join(DIRECTORY, 'illegaly:named:file')

  def setUp(self):
    test_utils.set_up_pyfakefs(self)
    self.fs.create_dir(self.DIRECTORY)
    self.fs.create_file(self.FILE_PATH, contents=self.FILE_CONTENTS)


class FileHashTest(FileMixin, fake_filesystem_unittest.TestCase):
  """Tests utils.file_hash works as expected."""

  def test_sha1sum(self):
    """Test that the correct sha1sum is calculated."""
    self.assertEqual(self.FILE_SHA1SUM, utils.file_hash(self.FILE_PATH))


class LegalizeFilenamesTest(FileMixin, fake_filesystem_unittest.TestCase):
  """Tests for legalize_filenames."""

  def test_rename_illegal(self):
    """Test that illegally named files are renamed."""
    legally_named = corpus_manager.legalize_filenames([self.FILE_PATH])
    # pylint: disable=unnecessary-comprehension
    self.assertEqual([os.path.join(self.DIRECTORY, self.FILE_SHA1SUM)],
                     [file_path for file_path in legally_named])
    with open(os.path.join(self.DIRECTORY, self.FILE_SHA1SUM)) as file_handle:
      self.assertEqual(self.FILE_CONTENTS, file_handle.read())

  def test_does_not_rename_legal(self):
    """Test that legally named files are not renamed."""
    new_file_path_1 = os.path.join(self.DIRECTORY, 'new_file')
    os.rename(self.FILE_PATH, new_file_path_1)
    initial_files = os.listdir(self.DIRECTORY)
    new_file_path_2 = '/other_new_file'
    initial_files.append(new_file_path_2)
    legal_files = corpus_manager.legalize_filenames(initial_files)
    self.assertEqual(initial_files, legal_files)
    with open(new_file_path_1) as file_handle:
      self.assertEqual(self.FILE_CONTENTS, file_handle.read())

  def test_logs_errors(self):
    """Test that errors are logged when we fail to rename a file."""
    test_helpers.patch(
        self, ['shutil.move', 'clusterfuzz._internal.metrics.logs.log_error'])

    def mock_move(*args, **kwargs):  # pylint: disable=unused-argument
      raise OSError

    self.mock.move.side_effect = mock_move
    legal_files = corpus_manager.legalize_filenames([self.FILE_PATH])
    self.assertEqual([], legal_files)
    failed_to_move_files = [(self.FILE_PATH,
                             os.path.join(self.DIRECTORY, self.FILE_SHA1SUM))]

    self.mock.log_error.assert_called_with(
        'Failed to rename files.', failed_to_move_files=failed_to_move_files)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generic helper functions useful in tests."""

import atexit
import datetime
import io
import os
import shutil
import socket
import subprocess
import tempfile
import threading
import unittest

import requests
import six

from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_init
from clusterfuzz._internal.google_cloud_utils import pubsub
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import process_handler

CURRENT_TIME = datetime.datetime.utcnow()
EMULATOR_TIMEOUT = 20

# Per-process emulator instances.
_emulators = {}


def create_generic_testcase(created_days_ago=28):
  """Create a simple test case."""
  testcase = data_types.Testcase()

  # Add more values here as needed. Intended to be the bare minimum for what we
  # need to simulate a test case.
  testcase.absolute_path = '/a/b/c/test.html'
  testcase.crash_address = '0xdeadbeef'
  testcase.crash_revision = 1
  testcase.crash_state = 'crashy_function()'
  testcase.crash_stacktrace = testcase.crash_state
  testcase.crash_type = 'fake type'
  testcase.comments = 'Fuzzer: test'
  testcase.fuzzed_keys = 'abcd'
  testcase.minimized_keys = 'efgh'
  testcase.fuzzer_name = 'fuzzer1'
  testcase.open = True
  testcase.one_time_crasher_flag = False
  testcase.job_type = 'test_content_shell_drt'
  testcase.status = 'Processed'
  testcase.timestamp = CURRENT_TIME - datetime.timedelta(days=created_days_ago)
  testcase.project_name = 'project'
  testcase.platform = 'linux'
  testcase.put()

  return testcase


def entities_equal(entity_1, entity_2, check_key=True):
  """Return a bool on whether two input entities are the same."""
  if check_key:
    return entity_1.key == entity_2.key

  return entity_1.to_dict() == entity_2.to_dict()


def entity_exists(entity):
  """Return a bool on where the entity exists in datastore."""
  return entity.get_by_id(entity.key.id())


def adhoc(func):
  """Mark the testcase as an adhoc. Adhoc tests are NOT expected to run before
    merging and are NOT counted toward test coverage; they are used to test
    tricky situations.

    Another way to think about it is that, if there was no adhoc test, we
    would write a Python script (which is not checked in) to test what we want
    anyway... so, it's better to check in the script.

    For example, downloading a chrome revision (10GB) and
    unpacking it. It can be enabled using the env ADHOC=1."""
  return unittest.skipIf(not environment.get_value('ADHOC', False),
                         'Adhoc tests are not enabled.')(
                             func)


def integration(func):
  """Mark the testcase as integration because it depends on network resources
    and/or is slow. The integration tests should, at least, be run before
    merging and are counted toward test coverage. It can be enabled using the
    env INTEGRATION=1."""
  return unittest.skipIf(not environment.get_value('INTEGRATION', False),
                         'Integration tests are not enabled.')(
                             func)


def slow(func):
  """Slow tests which are skipped during presubmit."""
  return unittest.skipIf(not environment.get_value('SLOW_TESTS', True),
                         'Skipping slow tests.')(
                             func)


def reproduce_tool(func):
  """Tests for the test case reproduction script."""
  return unittest.skipIf(
      not environment.get_value('REPRODUCE_TOOL_TESTS', False),
      'Skipping reproduce tool tests.')(
          func)


def android_device_required(func):
  """Skip Android-specific tests if we cannot run them."""
  reason = None
  if not environment.get_value('ANDROID_SERIAL'):
    reason = 'Android device tests require that ANDROID_SERIAL is set.'
  elif not environment.get_value('INTEGRATION'):
    reason = 'Integration tests are not enabled.'
  elif environment.platform() != 'LINUX':
    reason = 'Android device tests can only run on a Linux host.'

  return unittest.skipIf(reason is not None, reason)(func)


class EmulatorInstance(object):
  """Emulator instance."""

  def __init__(self, proc, port, read_thread, data_dir):
    self._proc = proc
    self._port = port
    self._read_thread = read_thread
    self._data_dir = data_dir

  def cleanup(self):
    """Stop and clean up the emulator."""
    process_handler.terminate_root_and_child_processes(self._proc.pid)
    self._read_thread.join()
    if self._data_dir:
      shutil.rmtree(self._data_dir, ignore_errors=True)

  def reset(self):
    """Reset emulator state."""
    req = requests.post('http://localhost:{}/reset'.format(self._port))
    req.raise_for_status()


def _find_free_port():
  """Find a free port."""
  sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  sock.bind(('localhost', 0))
  _, port = sock.getsockname()
  sock.close()

  return port


def wait_for_emulator_ready(proc,
                            emulator,
                            indicator,
                            timeout=EMULATOR_TIMEOUT,
                            output_lines=None):
  """Wait for emulator to be ready."""
  if output_lines is None:
    output_lines = []

  def _read_thread(proc, ready_event):
    """Thread to continuously read from the process stdout."""
    ready = False
    while True:
      line = proc.stdout.readline()
      if not line:
        break

      if output_lines is not None:
        output_lines.append(line)

      if not ready and indicator in line:
        ready = True
        ready_event.set()

  # Wait for process to become ready.
  ready_event = threading.Event()
  thread = threading.Thread(target=_read_thread, args=(proc, ready_event))
  thread.daemon = True
  thread.start()

  if not ready_event.wait(timeout):
    output = b'\n'.join(output_lines).decode()
    raise RuntimeError(
        f'{emulator} emulator did not get ready in time: {output}.')

  return thread


def start_cloud_emulator(emulator,
                         args=None,
                         data_dir=None,
                         store_on_disk=False):
  """Start a cloud emulator."""
  ready_indicators = {
      'datastore': b'is now running',
      'pubsub': b'Server started',
  }

  store_on_disk_flag = ('--store-on-disk'
                        if store_on_disk else '--no-store-on-disk')
  default_flags = {
      'datastore': [store_on_disk_flag, '--consistency=1'],
      'pubsub': [],
  }

  if emulator not in ready_indicators:
    raise RuntimeError('Unsupported emulator')

  if data_dir:
    cleanup_dir = None
  else:
    temp_dir = tempfile.mkdtemp()
    data_dir = temp_dir
    cleanup_dir = temp_dir

  port = _find_free_port()

  command = [
      'gcloud', 'beta', 'emulators', emulator, 'start',
      '--data-dir=' + data_dir, '--host-port=localhost:' + str(port),
      '--project=' + local_config.GAEConfig().get('application_id')
  ]
  if args:
    command.extend(args)

  command.extend(default_flags[emulator])

  # Start emulator.
  proc = subprocess.Popen(
      command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

  thread = wait_for_emulator_ready(proc, emulator, ready_indicators[emulator])

  # Set env vars.
  env_vars = subprocess.check_output([
      'gcloud', 'beta', 'emulators', emulator, 'env-init',
      '--data-dir=' + data_dir
  ])

  for line in env_vars.splitlines():
    key, value = line.split()[1].split(b'=')
    os.environ[key.strip().decode('utf-8')] = value.strip().decode('utf-8')

  return EmulatorInstance(proc, port, thread, cleanup_dir)


def create_pubsub_topic(client, project, name):
  """Create topic if it doesn't exist."""
  full_name = pubsub.topic_name(project, name)
  if client.get_topic(full_name):
    return

  client.create_topic(full_name)


def create_pubsub_subscription(client, project, topic, name):
  """Create subscription if it doesn't exist."""
  topic_name = pubsub.topic_name(project, topic)
  full_name = pubsub.subscription_name(project, name)
  if client.get_subscription(full_name):
    return

  client.create_subscription(full_name, topic_name)


def setup_pubsub(project):
  """Set up pubsub topics and subscriptions."""
  config = local_config.Config('pubsub.queues')
  client = pubsub.PubSubClient()

  queues = config.get('resources')

  for queue in queues:
    create_pubsub_topic(client, project, queue['name'])
    create_pubsub_subscription(client, project, queue['name'], queue['name'])


def with_cloud_emulators(*emulator_names):
  """Decorator for starting cloud emulators from a unittest.TestCase."""

  def decorator(cls):
    """Decorator."""

    class Wrapped(cls):
      """Wrapped class."""

      @classmethod
      def setUpClass(cls):
        """Class setup."""
        for emulator_name in emulator_names:
          if emulator_name not in _emulators:
            _emulators[emulator_name] = start_cloud_emulator(emulator_name)
            atexit.register(_emulators[emulator_name].cleanup)

          if emulator_name == 'datastore':
            cls._context_generator = ndb_init.context()
            cls._context_generator.__enter__()

        super(Wrapped, cls).setUpClass()

      @classmethod
      def tearDownClass(cls):
        """Class teardown."""
        for emulator_name in emulator_names:
          if emulator_name == 'datastore':
            cls._context_generator.__exit__(None, None, None)

        super(Wrapped, cls).tearDownClass()

      def setUp(self):
        for emulator in six.itervalues(_emulators):
          emulator.reset()

        super().setUp()

    Wrapped.__module__ = cls.__module__
    Wrapped.__name__ = cls.__name__
    return Wrapped

  return decorator


def set_up_pyfakefs(test_self, allow_root_user=True):
  """Helper to set up Pyfakefs."""
  real_cwd = os.path.realpath(os.getcwd())
  config_dir = os.path.realpath(environment.get_config_directory())
  test_self.setUpPyfakefs(allow_root_user=allow_root_user)
  test_self.fs.add_real_directory(config_dir, lazy_read=False)
  os.chdir(real_cwd)


def supported_platforms(*platforms):
  """Decorator for enabling tests only on certain platforms."""

  def decorator(func):  # pylint: disable=unused-argument
    """Decorator."""
    return unittest.skipIf(environment.platform() not in platforms,
                           'Unsupported platform.')(
                               func)

  return decorator


MockStdout = io.StringIO  # pylint: disable=invalid-name



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Google Cloud Profiler integration."""

import time
import unittest

from clusterfuzz._internal.metrics import profiler
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers


class ProfilerTest(unittest.TestCase):
  """Test Google Cloud Profiler."""

  def setUp(self):
    helpers.patch_environ(self)
    environment.set_value('USE_PYTHON_PROFILER', True)

    self.fake_profile = {
        'profileType': 'type',
        'deployment': 'deployment',
        'duration': 123,
    }

    helpers.patch(self, [
        'googlecloudprofiler.backoff.Backoff.next_backoff',
        'googlecloudprofiler.client.Client.setup_auth',
        'googlecloudprofiler.client.Client._build_service',
        'googlecloudprofiler.client.Client._create_profile',
        'googlecloudprofiler.client.Client._collect_and_upload_profile',
    ])

    # Time in seconds used in a sleep call inside profiler loop.
    self.mock.next_backoff.return_value = 0.005
    self.mock.setup_auth.return_value = 'project_id'
    self.mock._build_service.return_value = None  # pylint: disable=protected-access
    self.mock._create_profile.return_value = self.fake_profile  # pylint: disable=protected-access

  def test_profiler(self):
    """Test profiler."""
    profiler.start_if_needed('python_profiler_unit_test_service')

    # A dummy code to spend a few moments on.
    counter = 12345
    for _ in range(5):
      time.sleep(0.01)
      counter *= counter

    self.assertLess(0,
                    len(self.mock._collect_and_upload_profile.call_args_list))  # pylint: disable=protected-access
    self.assertEqual((self.fake_profile,),
                     self.mock._collect_and_upload_profile.call_args[0][1:])  # pylint: disable=protected-access



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Bisection infrastructure functions."""

from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.build_management import revisions
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.crash_analysis import severity_analyzer
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.google_cloud_utils import blobs
from clusterfuzz._internal.google_cloud_utils import pubsub
from clusterfuzz._internal.system import environment


def _get_topic():
  """Get the Pub/Sub topic for publishing tasks."""
  return local_config.ProjectConfig().get('bisect_service.pubsub_topic')


def notify_bisection_invalid(testcase):
  """Notify the bisection infrastructure of a testcase getting into invalid
  state."""
  pubsub_topic = _get_topic()
  if not pubsub_topic:
    return

  pubsub_client = pubsub.PubSubClient()
  pubsub_client.publish(pubsub_topic, [
      pubsub.Message(b'', {
          'type': 'invalid',
          'testcase_id': str(testcase.key.id()),
      })
  ])


def request_bisection(testcase):
  """Request precise bisection."""
  pubsub_topic = _get_topic()
  if not pubsub_topic:
    return

  # Only request bisects for reproducible security bugs with a bug filed, found
  # by engine fuzzers.
  if not testcase.security_flag:
    return

  if testcase.fixed == 'NA':
    # Testcase got into an invalid state.
    notify_bisection_invalid(testcase)
    return

  if testcase.one_time_crasher_flag:
    return

  if not testcase.bug_information:
    return

  target = testcase.get_fuzz_target()
  if not target:
    return

  # Only make 1 request of each type per testcase.
  if (not testcase.get_metadata('requested_regressed_bisect') and
      _make_bisection_request(pubsub_topic, testcase, target, 'regressed')):
    testcase.set_metadata('requested_regressed_bisect', True)

  if (not testcase.get_metadata('requested_fixed_bisect') and
      _make_bisection_request(pubsub_topic, testcase, target, 'fixed')):
    testcase.set_metadata('requested_fixed_bisect', True)


def _check_commits(testcase, bisect_type, old_commit, new_commit):
  """Check old and new commit validity."""
  if old_commit != new_commit or build_manager.is_custom_binary():
    return old_commit, new_commit

  # Something went wrong during bisection for the same commit to be chosen for
  # both the start and end range.
  # Get the bisection infrastructure to re-bisect.
  if environment.is_running_on_app_engine():
    bucket_path = data_handler.get_value_from_job_definition(
        testcase.job_type, 'RELEASE_BUILD_BUCKET_PATH')
  else:
    bucket_path = build_manager.get_primary_bucket_path()
  revision_list = build_manager.get_revisions_list(bucket_path)

  last_tested_revision = testcase.get_metadata('last_tested_crash_revision')
  known_crash_revision = last_tested_revision or testcase.crash_revision

  if bisect_type == 'fixed':
    # Narrowest range: last crashing revision up to the latest build.
    return _get_commits(
        str(known_crash_revision) + ':' + str(revision_list[-1]),
        testcase.job_type)

  if bisect_type == 'regressed':
    # Narrowest range: first build to the first crashing revision.
    return _get_commits(
        str(revision_list[0]) + ':' + str(testcase.crash_revision),
        testcase.job_type)

  raise ValueError('Invalid bisection type: ' + bisect_type)


def _make_bisection_request(pubsub_topic, testcase, target, bisect_type):
  """Make a bisection request to the external bisection service. Returns whether
  or not a request was actually made."""
  if bisect_type == 'fixed':
    old_commit, new_commit = _get_commits(testcase.fixed, testcase.job_type)
  elif bisect_type == 'regressed':
    old_commit, new_commit = _get_commits(testcase.regression,
                                          testcase.job_type)
  else:
    raise ValueError('Invalid bisection type: ' + bisect_type)

  if not new_commit:
    # old_commit can be empty (i.e. '0' case), but new_commit should never be.
    return False

  old_commit, new_commit = _check_commits(testcase, bisect_type, old_commit,
                                          new_commit)

  reproducer = blobs.read_key(testcase.minimized_keys or testcase.fuzzed_keys)
  pubsub_client = pubsub.PubSubClient()
  pubsub_client.publish(pubsub_topic, [
      pubsub.Message(
          reproducer, {
              'type':
                  bisect_type,
              'project_name':
                  target.project,
              'sanitizer':
                  environment.SANITIZER_NAME_MAP[
                      environment.get_memory_tool_name(testcase.job_type)
                  ],
              'fuzz_target':
                  target.binary,
              'old_commit':
                  old_commit,
              'new_commit':
                  new_commit,
              'testcase_id':
                  str(testcase.key.id()),
              'issue_id':
                  testcase.bug_information,
              'crash_type':
                  testcase.crash_type,
              'crash_state':
                  testcase.crash_state,
              'security':
                  str(testcase.security_flag),
              'severity':
                  severity_analyzer.severity_to_string(
                      testcase.security_severity),
              'timestamp':
                  testcase.timestamp.isoformat(),
          })
  ])
  return True


def _get_commits(commit_range, job_type):
  """Get commits from range."""
  if not commit_range or commit_range == 'NA':
    return None, None

  start, end = revisions.get_start_and_end_revision(commit_range)
  components = revisions.get_component_range_list(start, end, job_type)
  if not components:
    return None, None

  commits = components[0]['link_text']

  if ':' not in commits:
    return commits, commits

  old_commit, new_commit = commits.split(':')
  if old_commit == '0':
    old_commit = ''

  return old_commit, new_commit



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Process runner."""

import subprocess


class Process(object):
  """Represent an child process.

       This class is intentionally similar to subprocess, except that it allows
       various fields to be set before executing the process. Additionally, it
       allows tests to overload process creation and execution in one place;
       see MockProcess.
    """

  def __init__(self, args, **kwargs):
    self.args = args
    self.cwd = kwargs.get('cwd', None)
    self.stdin = kwargs.get('stdin', None)
    self.stdout = kwargs.get('stdout', None)
    self.stderr = kwargs.get('stderr', None)

  def popen(self):
    """Analogous to subprocess.Popen."""
    p = subprocess.Popen(
        self.args, stdin=self.stdin, stdout=self.stdout, stderr=self.stderr)
    self.__init__([])
    return p

  def call(self):
    """Analogous to subprocess.call."""
    p = self.popen()
    return p.wait()

  def check_call(self):
    """Analogous to subprocess.check_call."""
    cmd = self.args
    rc = self.call()
    if rc != 0:
      raise subprocess.CalledProcessError(rc, cmd)
    return rc

  def check_output(self):
    """Analogous to subprocess.check_output."""
    cmd = self.args
    self.stdout = subprocess.PIPE
    p = self.popen()
    out, _ = p.communicate()
    rc = p.returncode
    if rc != 0:
      raise subprocess.CalledProcessError(rc, cmd)
    return out



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test helpers for Android."""

import unittest

from clusterfuzz._internal.platforms.android import adb
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


@test_utils.android_device_required
class AndroidTest(unittest.TestCase):
  """Set up state for Android tests."""

  def setUp(self):
    helpers.patch_environ(self)
    environment.set_value('OS_OVERRIDE', 'ANDROID')
    environment.set_bot_environment()
    adb.setup_adb()
    adb.run_as_root()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generate Syzkaller fuzzer config dynamically."""

import json
import os


def generate(serial,
             work_dir_path,
             binary_path,
             vmlinux_path,
             config_path,
             kcov=True,
             reproduce=True,
             syzhub_address=None,
             syzhub_client=None,
             syzhub_key=None,
             on_cuttlefish=False):
  """Generates syzkaller config file.

  Args:
    serial: (str)serial number of the device being fuzzed.
    work_dir_path: (str) path to working directory of syzkaller.
    binary_path: (str) path to syzkaller binaries.
    vmlinux_path: (str) path to the vmlinux file.
    config_path: (str) path to write the syzkaller config to.
    kcov: (boolean) true if coverage is enabled.
    reproduce: (boolean) true if repro is enabled.
    syzhub_address: (str) ip:host of the syzhub to connect to.
    syzhub_client: (str) syzhub client name.
    syzhub_key: (str) syzhub key.
  """
  devices = {}
  devices['devices'] = [serial]
  data = {}
  data['target'] = 'linux/arm64'
  data['reproduce'] = reproduce
  data['workdir'] = work_dir_path
  data['http'] = 'localhost:0'
  data['syzkaller'] = binary_path
  #TODO(hzawawy): consider what suppressions are best for Android.
  data['suppressions'] = ['do_rt_sigqueueinfo', 'do_rt_tgsigqueueinfo']
  data['vm'] = devices
  data['kernel_obj'] = vmlinux_path
  data['sandbox'] = 'android'
  data['ignores'] = ['WARNING:', 'INFO:']
  data['type'] = 'adb'
  data['procs'] = 1
  data['cover'] = kcov

  if on_cuttlefish:
    data['target'] = 'linux/amd64'
    data['disable_syscalls'] = ['openat$vhost_vsock']
    data['sandbox'] = 'none'

  if syzhub_address and syzhub_client and syzhub_key:
    data['hub_addr'] = syzhub_address
    data['hub_client'] = syzhub_client
    data['hub_key'] = syzhub_key
    data['name'] = '{}-{}'.format(syzhub_client, serial)

  ensure_dir(config_path)
  with open(config_path, 'w') as write_file:
    json.dump(data, write_file)


def ensure_dir(file_path):
  directory = os.path.dirname(file_path)
  if not os.path.exists(directory):
    os.makedirs(directory)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for server module."""
import unittest

from clusterfuzz._internal.tests.test_libs import helpers


class ServerTest(unittest.TestCase):
  """Test server module is loaded."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.metrics.logs._is_running_on_app_engine',
    ])
    self.mock._is_running_on_app_engine.return_value = True  # pylint: disable=protected-access

  # pylint: disable=protected-access
  def test(self):
    import server
    self.assertIsNotNone(server.handlers)
    self.assertIsNotNone(server.cron_routes)
    self.assertIsNotNone(server.app)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""corpus backup handler."""

import datetime
import os

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import fuzz_target_utils
from clusterfuzz._internal.datastore import ndb_utils
from clusterfuzz._internal.fuzzing import corpus_manager
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from handlers import base_handler
from libs import handler


def _set_public_acl_if_needed(url):
  """Sets public ACL on the object with given URL, if it's not public yet."""
  if storage.get_acl(url, 'allUsers'):
    logs.log('%s is already marked public, skipping.' % url)
    return True

  if not storage.set_acl(url, 'allUsers'):
    logs.log_error('Failed to mark %s public.' % url)
    return False

  return True


def _make_corpus_backup_public(target, corpus_fuzzer_name_override,
                               corpus_backup_bucket_name):
  """Identifies old corpus backups and makes them public."""
  corpus_backup_date = utils.utcnow().date() - datetime.timedelta(
      days=data_types.CORPUS_BACKUP_PUBLIC_LOOKBACK_DAYS)

  corpus_backup_url = corpus_manager.gcs_url_for_backup_file(
      corpus_backup_bucket_name, corpus_fuzzer_name_override or target.engine,
      target.project_qualified_name(), corpus_backup_date)

  if not storage.get(corpus_backup_url):
    logs.log_warn('Failed to find corpus backup %s.' % corpus_backup_url)
    return

  if not _set_public_acl_if_needed(corpus_backup_url):
    return

  filename = (
      corpus_manager.PUBLIC_BACKUP_TIMESTAMP + os.extsep +
      corpus_manager.BACKUP_ARCHIVE_FORMAT)
  public_url = os.path.join(os.path.dirname(corpus_backup_url), filename)

  if not storage.copy_blob(corpus_backup_url, public_url):
    logs.log_error(
        'Failed to overwrite %s with the latest public corpus backup.' %
        public_url)
    return

  if not _set_public_acl_if_needed(public_url):
    return

  logs.log('Corpus backup %s is now marked public.' % corpus_backup_url)


class MakePublicHandler(base_handler.Handler):
  """Makes corpuses older than 90 days public."""

  @handler.cron()
  def get(self):
    """Handle a GET request."""
    jobs = ndb_utils.get_all_from_model(data_types.Job)
    default_backup_bucket = utils.default_backup_bucket()
    for job in jobs:
      job_environment = job.get_environment()
      if utils.string_is_true(job_environment.get('EXPERIMENTAL')):
        # Don't use corpus backups from experimental jobs. Skip.
        continue

      if not utils.string_is_true(job_environment.get('CORPUS_PRUNE')):
        # There won't be any corpus backups for these jobs. Skip.
        continue

      corpus_backup_bucket_name = job_environment.get('BACKUP_BUCKET',
                                                      default_backup_bucket)
      if not corpus_backup_bucket_name:
        # No backup bucket found. Skip.
        continue

      corpus_fuzzer_name_override = job_environment.get(
          'CORPUS_FUZZER_NAME_OVERRIDE')

      target_jobs = list(fuzz_target_utils.get_fuzz_target_jobs(job=job.name))
      fuzz_targets = fuzz_target_utils.get_fuzz_targets_for_target_jobs(
          target_jobs)

      for target in fuzz_targets:
        if not target:
          # This is expected if any fuzzer/job combinations become outdated.
          continue

        try:
          _make_corpus_backup_public(target, corpus_fuzzer_name_override,
                                     corpus_backup_bucket_name)
        except:
          logs.log_error('Failed to make %s corpus backup public.' % target)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers used to generate Content Security Policies for pages."""
import collections

from libs import auth


class CSPBuilder(object):
  """Helper to build a Content Security Policy string."""

  def __init__(self):
    self.directives = collections.defaultdict(list)

  def add(self, directive, source, quote=False):
    """Add a source for a given directive."""
    # Some values for sources are expected to be quoted. No escaping is done
    # since these are specific literal values that don't require it.
    if quote:
      source = '\'{}\''.format(source)

    assert source not in self.directives[directive], (
        'Duplicate source "{source}" for directive "{directive}"'.format(
            source=source, directive=directive))
    self.directives[directive].append(source)

  def add_sourceless(self, directive):
    assert directive not in self.directives, (
        'Sourceless directive "{directive}" already exists.'.format(
            directive=directive))

    self.directives[directive] = []

  def remove(self, directive, source, quote=False):
    """Remove a source for a given directive."""
    if quote:
      source = '\'{}\''.format(source)

    assert source in self.directives[directive], (
        'Removing nonexistent "{source}" for directive "{directive}"'.format(
            source=source, directive=directive))
    self.directives[directive].remove(source)

  def __str__(self):
    """Convert to a string to send with a Content-Security-Policy header."""
    parts = []

    # Sort directives for deterministic results.
    for directive, sources in sorted(self.directives.items()):
      # Each policy part has the form "directive source1 source2 ...;".
      parts.append(' '.join([directive] + sources) + ';')

    return ' '.join(parts)


def get_default_builder():
  """Get a CSPBuilder object for the default policy.

  Can be modified for specific pages if needed."""
  builder = CSPBuilder()

  # By default, disallow everything. Whitelist only features that are needed.
  builder.add('default-src', 'none', quote=True)

  # Allow various directives if sourced from self.
  builder.add('font-src', 'self', quote=True)
  builder.add('connect-src', 'self', quote=True)
  builder.add('img-src', 'self', quote=True)
  builder.add('manifest-src', 'self', quote=True)

  # External scripts. Google analytics, charting libraries.
  builder.add('script-src', 'www.google-analytics.com')
  builder.add('script-src', 'www.gstatic.com')
  builder.add('script-src', 'apis.google.com')

  # Google Analytics also uses connect-src and img-src.
  builder.add('connect-src', 'www.google-analytics.com')
  builder.add('img-src', 'www.google-analytics.com')

  # Firebase.
  builder.add('img-src', 'www.gstatic.com')
  builder.add('connect-src', 'securetoken.googleapis.com')
  builder.add('connect-src', 'www.googleapis.com')
  builder.add('frame-src', auth.auth_domain())

  # External style. Used for fonts, charting libraries.
  builder.add('style-src', 'fonts.googleapis.com')
  builder.add('style-src', 'www.gstatic.com')

  # External fonts.
  builder.add('font-src', 'fonts.gstatic.com')

  # Some upload forms require us to connect to the cloud storage API.
  builder.add('connect-src', 'storage.googleapis.com')

  # Mixed content is unexpected, but upgrade requests rather than block.
  builder.add_sourceless('upgrade-insecure-requests')

  # We don't expect object to be used, but it doesn't fall back to default-src.
  builder.add('object-src', 'none', quote=True)

  # We don't expect workers to be used, but they fall back to script-src.
  builder.add('worker-src', 'none', quote=True)

  # Add reporting so that violations don't break things silently.
  builder.add('report-uri', '/report-csp-failure')

  # TODO(mbarbella): Remove Google-specific cases by allowing configuration.

  # Internal authentication.
  builder.add('manifest-src', 'login.corp.google.com')

  # TODO(mbarbella): Improve the policy by limiting the additions below.

  # Because we use Polymer Bundler to create large files containing all of our
  # scripts inline, our policy requires this (which weakens CSP significantly).
  builder.add('script-src', 'unsafe-inline', quote=True)

  # Some of the pages that read responses from json handlers require this.
  builder.add('script-src', 'unsafe-eval', quote=True)

  # Our Polymer Bundler usage also requires inline style.
  builder.add('style-src', 'unsafe-inline', quote=True)

  # Some fonts are loaded from data URIs.
  builder.add('font-src', 'data:')

  return builder


def get_default():
  """Get the default Content Security Policy as a string."""
  return str(get_default_builder())
